# coding: utf-8
"""

"""

import enum
import typing
import datetime

from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .authentication import Authentication
    from .components import Component
    from .groups import Group
    from .user_log import UserLogEntry


@enum.unique
class UserType(enum.Enum):
    PERSON = 1
    OTHER = 2
    FEDERATION_USER = 3
    ELN_IMPORT_USER = 4


class User(Model):
    __tablename__ = 'users'
    __table_args__ = (
        db.CheckConstraint(
            '(type = \'FEDERATION_USER\' AND NOT is_admin AND fed_id IS NOT NULL AND component_id IS NOT NULL) OR (type = \'ELN_IMPORT_USER\' AND NOT is_admin AND eln_import_id IS NOT NULL AND eln_object_id IS NOT NULL) OR (name IS NOT NULL AND email IS NOT NULL)',
            name='users_not_null_check'
        ),
        db.UniqueConstraint('fed_id', 'component_id', name='users_fed_id_component_id_key'),
        db.UniqueConstraint('eln_import_id', 'eln_object_id', name='users_eln_import_id_eln_object_id_key'),
    )

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    email: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    type: Mapped[UserType] = db.Column(db.Enum(UserType), nullable=False)
    is_admin: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    is_readonly: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    is_hidden: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    orcid: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    affiliation: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    is_active: Mapped[bool] = db.Column(db.Boolean, default=True, nullable=False)
    role: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    extra_fields: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False, default={}, server_default=db.text("'{}'::json"))
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    last_modified: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    last_modified_by_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
    # named foreign key constraint to allow for manual deletion in empty_database()
    eln_import_id = db.Column(db.Integer, db.ForeignKey('eln_imports.id', use_alter=True, name='fk_users_eln_import_id'), nullable=True)
    eln_object_id = db.Column(db.String, nullable=True)
    eln_import = db.relationship('ELNImport', foreign_keys=[eln_import_id])
    groups: Mapped[typing.List['Group']] = relationship('Group', secondary='user_group_memberships', back_populates='members')
    log_entries: Mapped[typing.List['UserLogEntry']] = relationship('UserLogEntry', back_populates="user")
    authentication_methods: Mapped[typing.List['Authentication']] = relationship('Authentication', back_populates='user')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["User"]]

    def __init__(
            self,
            name: typing.Optional[str],
            email: typing.Optional[str],
            type: UserType,
            orcid: typing.Optional[str] = None,
            affiliation: typing.Optional[str] = None,
            role: typing.Optional[str] = None,
            extra_fields: typing.Optional[typing.Dict[str, str]] = None,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None,
            last_modified: typing.Optional[datetime.datetime] = None,
            *,
            eln_import_id: typing.Optional[int] = None,
            eln_object_id: typing.Optional[str] = None,
    ) -> None:
        super().__init__(
            name=name,
            email=email,
            type=type,
            orcid=orcid,
            affiliation=affiliation,
            role=role,
            extra_fields=extra_fields if extra_fields is not None else {},
            fed_id=fed_id,
            component_id=component_id,
            last_modified=last_modified if last_modified is not None else datetime.datetime.now(datetime.timezone.utc),
            eln_import_id=eln_import_id,
            eln_object_id=eln_object_id
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, User):
            return bool(
                self.id == other.id and
                self.name == other.name and
                self.email == other.email and
                self.type == other.type and
                self.is_admin == other.is_admin and
                self.is_readonly == other.is_readonly and
                self.is_hidden == other.is_hidden and
                self.orcid == other.orcid and
                self.affiliation == other.affiliation and
                self.role == other.role and
                self.fed_id == other.fed_id and
                self.component_id == other.component_id and
                self.eln_import_id == other.eln_import_id and
                self.eln_object_id == other.eln_object_id
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, name={self.name})>'


class UserInvitation(Model):
    __tablename__ = 'user_invitations'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    inviter_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    accepted: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserInvitation"]]


class UserFederationAlias(Model):
    __tablename__ = 'fed_user_aliases'

    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    name: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    use_real_name: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    email: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    use_real_email: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    orcid: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    use_real_orcid: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    affiliation: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    use_real_affiliation: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    role: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    use_real_role: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    extra_fields: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False, default={}, server_default=db.text("'{}'::json"))
    last_modified: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    user: Mapped[User] = relationship('User')
    component: Mapped['Component'] = relationship('Component')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserFederationAlias"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(user_id, component_id),
    )

    def __init__(
            self,
            user_id: int,
            component_id: int,
            name: typing.Optional[str] = None,
            use_real_name: bool = False,
            email: typing.Optional[str] = None,
            use_real_email: bool = False,
            orcid: typing.Optional[str] = None,
            use_real_orcid: bool = False,
            affiliation: typing.Optional[str] = None,
            use_real_affiliation: bool = False,
            role: typing.Optional[str] = None,
            use_real_role: bool = False,
            extra_fields: typing.Optional[typing.Dict[str, str]] = None,
            last_modified: typing.Optional[datetime.datetime] = None):
        super().__init__(
            user_id=user_id,
            component_id=component_id,
            name=name,
            use_real_name=use_real_name,
            email=email,
            use_real_email=use_real_email,
            orcid=orcid,
            use_real_orcid=use_real_orcid,
            affiliation=affiliation,
            use_real_affiliation=use_real_affiliation,
            role=role,
            use_real_role=use_real_role,
            extra_fields=extra_fields if extra_fields is not None else {},
            last_modified=last_modified if last_modified is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(user_id={self.user_id}, component_id={self.component_id}; name={self.name})>'


class FederatedIdentity(Model):
    __tablename__ = 'fed_identities'

    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False, primary_key=True)
    user: Mapped[User] = relationship('User', foreign_keys=[user_id])
    local_fed_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False, primary_key=True)
    local_fed_user: Mapped[User] = relationship('User', foreign_keys=[local_fed_id])
    active: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    login: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FederatedIdentity"]]

    def __init__(self, user_id: int, local_fed_id: int, active: bool = True, login: bool = False):
        super().__init__(user_id=user_id, local_fed_id=local_fed_id, active=active, login=login)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(user_id={self.user_id}, local_fed_id={self.local_fed_id}, active={self.active}, login={self.login})>"



# coding: utf-8
"""
Add migration_indices table to database.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    migration_index_exists = db.session.execute(db.text("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'migration_index'
    """)).first() is not None
    if migration_index_exists:
        return False

    # Perform migration
    db.session.execute(db.text("""
        CREATE TABLE migration_index (
          migration_index INTEGER
        )
    """))

    # Insert migration index 0
    db.session.execute(db.text("""
        INSERT INTO migration_index
        (migration_index) VALUES (0)
    """))
    return True



# coding: utf-8
"""
RESTful API for SampleDB
"""
import typing

import flask

from .authentication import multi_auth
from ..utils import Resource, ResponseData
from ...logic import errors, projects


def project_to_json(project: projects.Project) -> typing.Dict[str, typing.Any]:
    return {
        'project_id': project.id,
        'name': project.name,
        'description': project.description,
        'member_users': sorted(
            [
                {
                    'user_id': user_id,
                    'permissions': permissions.name.lower(),
                    'href': flask.url_for('api.user', user_id=user_id, _external=True),
                } for user_id, permissions in projects.get_project_member_user_ids_and_permissions(project_id=project.id).items()
            ],
            key=lambda member_user: member_user['user_id']
        ),
        'member_groups': sorted(
            [
                {
                    'group_id': group_id,
                    'permissions': permissions.name.lower(),
                    'href': flask.url_for('api.group', group_id=group_id, _external=True),
                }
                for group_id, permissions in projects.get_project_member_group_ids_and_permissions(project_id=project.id).items()
            ],
            key=lambda member_group: member_group['group_id']
        )
    }


class Projects(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        return [
            project_to_json(project)
            for project in sorted(projects.get_projects(), key=lambda project: project.id)
        ]


class Project(Resource):
    @multi_auth.login_required
    def get(self, project_id: int) -> ResponseData:
        try:
            project = projects.get_project(project_id=project_id)
        except errors.ProjectDoesNotExistError:
            return {
                "message": f"project {project_id} does not exist"
            }, 404
        return project_to_json(project)



# coding: utf-8
"""

"""

import base64
import datetime
import json
import typing

import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms.fields import StringField, HiddenField
from wtforms.validators import InputRequired
import pyotp
from fido2.webauthn import PublicKeyCredentialUserEntity, ResidentKeyRequirement, AttestedCredentialData

from .. import frontend
from .authentication import complete_sign_in
from ...logic import errors, authentication
from ...logic.users import get_user, User
from .forms import WebAuthnLoginForm
from ..utils import generate_qrcode
from ...utils import FlaskResponseT


class TOTPForm(FlaskForm):
    code = StringField(validators=[InputRequired()])
    description = StringField()


class WebauthnSetupForm(FlaskForm):
    fido2_passkey_credentials = HiddenField(validators=[InputRequired()])
    description = StringField(validators=[InputRequired()])


@frontend.route('/users/me/two_factor_authentication/totp/setup', methods=['GET', 'POST'])
@flask_login.login_required
def setup_totp_two_factor_authentication() -> FlaskResponseT:
    setup_form = TOTPForm()
    email = flask_login.current_user.email
    service_name = flask.current_app.config['SERVICE_NAME']
    if setup_form.validate_on_submit() and 'totp_secret' in flask.session:
        secret = flask.session['totp_secret']
        if pyotp.TOTP(secret).verify(setup_form.code.data):
            flask.session.pop('totp_secret')
            method = authentication.create_totp_two_factor_authentication_method(flask_login.current_user.id, secret, setup_form.description.data)
            authentication.activate_two_factor_authentication_method(method.id)
            flask.flash(_('Successfully set up two-factor authentication.'), 'success')
            return flask.redirect(flask.url_for('.user_me_preferences'))
        else:
            flask.flash(_('The code was invalid, please try again.'), 'error')
    else:
        secret = pyotp.random_base32()
        flask.session['totp_secret'] = secret
    url = pyotp.totp.TOTP(secret).provisioning_uri(name=email, issuer_name=service_name)
    qrcode_url = generate_qrcode(url, should_cache=False)
    return flask.render_template(
        'two_factor_authentication/set_up_totp.html',
        setup_form=setup_form,
        qrcode_url=qrcode_url,
        secret=secret
    ), 200, {
        'Cache-Control': 'no-cache, no-store, must-revalidate',
        'Pragma': 'no-cache',
        'Expires': '0'
    }


def _parse_confirm_data() -> typing.Optional[typing.Tuple[User, str, bool, bool, bool, typing.Optional[str]]]:
    confirm_data = flask.session.get('confirm_data')
    if confirm_data is None:
        return None
    try:
        expiration_datetime = datetime.datetime.strptime(confirm_data['expiration_datetime'], '%Y-%m-%d %H:%M:%S').replace(tzinfo=datetime.timezone.utc)
    except Exception:
        flask.session.pop('confirm_data')
        return None
    if expiration_datetime < datetime.datetime.now(datetime.timezone.utc):
        flask.session.pop('confirm_data')
        return None
    try:
        user = get_user(confirm_data['user_id'])
    except errors.UserDoesNotExistError:
        flask.session.pop('confirm_data')
        return None
    return user, confirm_data.get('reason'), confirm_data.get('is_for_refresh'), confirm_data.get('remember_me'), confirm_data.get('shared_device'), confirm_data.get('next_url')


@frontend.route('/users/me/two_factor_authentication/totp/confirm', methods=['GET', 'POST'])
def confirm_totp_two_factor_authentication() -> FlaskResponseT:
    confirm_data = _parse_confirm_data()
    if confirm_data is None:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        return flask.redirect(flask.url_for('.index'))
    user, reason, is_for_refresh, remember_me, shared_device, next_url = confirm_data
    if reason not in (
        'login',
        'activate_two_factor_authentication_method',
        'deactivate_two_factor_authentication_method'
    ):
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if reason in ('login', ) and flask_login.current_user.is_authenticated != is_for_refresh:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if reason in (
        'activate_two_factor_authentication_method',
        'deactivate_two_factor_authentication_method'
    ) and not flask_login.current_user.is_authenticated:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    all_methods = authentication.get_two_factor_authentication_methods(
        user_id=user.id,
        active=reason in (
            'login',
            'deactivate_two_factor_authentication_method'
        )
    )
    try:
        method_id = int(flask.request.args.get('method_id', ''))
    except ValueError:
        method_id = -1
    method = {
        method.id: method
        for method in all_methods
    }.get(method_id)
    if method is None:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if reason in (
        'deactivate_two_factor_authentication_method',
    ):
        method_to_deactivate_id = flask.session['confirm_data']['method_to_deactivate_id']
        method_to_deactivate = {
            method.id: method
            for method in all_methods
        }.get(method_to_deactivate_id)
        if method_to_deactivate is None:
            flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
            flask.session.pop('confirm_data')
            return flask.redirect(flask.url_for('.index'))
        if not method_to_deactivate.active:
            flask.flash(_('The two-factor authentication method has already been disabled.'), 'success')
            return flask.redirect(flask.url_for('.user_me_preferences'))
    else:
        method_to_deactivate = None

    if method.data.get('type') != 'totp':
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if not method.data.get('secret'):
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    confirm_form = TOTPForm()
    if confirm_form.validate_on_submit():
        secret = method.data['secret']
        if pyotp.TOTP(secret).verify(confirm_form.code.data):
            flask.session.pop('confirm_data')
            if reason == 'login':
                return complete_sign_in(user, is_for_refresh, remember_me, shared_device, next_url=next_url)
            elif reason == 'activate_two_factor_authentication_method':
                authentication.activate_two_factor_authentication_method(method.id)
                flask.flash(_('The two-factor authentication method has been enabled.'), 'success')
                return flask.redirect(flask.url_for('.user_me_preferences'))
            elif reason == 'deactivate_two_factor_authentication_method':
                assert method_to_deactivate is not None
                authentication.deactivate_two_factor_authentication_method(method_to_deactivate.id)
                flask.flash(_('The two-factor authentication method has been disabled.'), 'success')
                return flask.redirect(flask.url_for('.user_me_preferences'))
            else:
                return flask.abort(500)
        else:
            flask.flash(_('The code was invalid, please try again.'), 'error')
    return flask.render_template(
        'two_factor_authentication/confirm_totp.html',
        confirm_form=confirm_form,
        method_id=method_id,
        user=user,
        description=method.data.get('description'),
        reason=reason,
        method_to_deactivate=method_to_deactivate
    )


@frontend.route('/users/me/two_factor_authentication/fido2_passkey/setup', methods=['GET', 'POST'])
@flask_login.login_required
def setup_fido2_passkey_two_factor_authentication() -> FlaskResponseT:
    setup_form = WebauthnSetupForm()

    server = authentication.get_webauthn_server()
    options, state = server.register_begin(
        PublicKeyCredentialUserEntity(
            id=f"2fa-{flask_login.current_user.id}".encode('utf-8'),
            name=f"{flask_login.current_user.name} ({flask_login.current_user.email})",
            display_name=flask_login.current_user.name,
        ),
        credentials=authentication.get_all_two_factor_fido2_passkey_credentials_for_user(user_id=flask_login.current_user.id),
        resident_key_requirement=ResidentKeyRequirement.PREFERRED,
    )
    if setup_form.validate_on_submit():
        try:
            auth_data = server.register_complete(
                flask.session["webauthn_enroll_state"],
                json.loads(setup_form.fido2_passkey_credentials.data)
            )
            assert auth_data.credential_data is not None
            method = authentication.create_fido2_passkey_two_factor_authentication_method(
                user_id=flask_login.current_user.id,
                credential_data=auth_data.credential_data,
                description=setup_form.description.data
            )
            authentication.activate_two_factor_authentication_method(method.id)
            del flask.session["webauthn_enroll_state"]
            flask.flash(_('Successfully set up two-factor authentication.'), 'success')
        except Exception:
            flask.flash(_('Failed to set up two-factor authentication.'), 'error')
        return flask.redirect(flask.url_for('.user_me_preferences'))
    else:
        flask.session["webauthn_enroll_state"] = state

    return flask.render_template(
        'two_factor_authentication/set_up_fido2_passkey.html',
        setup_form=setup_form,
        options=dict(options),
    ), 200, {
        'Cache-Control': 'no-cache, no-store, must-revalidate',
        'Pragma': 'no-cache',
        'Expires': '0'
    }


@frontend.route('/users/me/two_factor_authentication/fido2_passkey/confirm', methods=['GET', 'POST'])
def confirm_fido2_passkey_two_factor_authentication() -> FlaskResponseT:
    confirm_data = _parse_confirm_data()
    if confirm_data is None:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        return flask.redirect(flask.url_for('.index'))
    user, reason, is_for_refresh, remember_me, shared_device, next_url = confirm_data
    if reason not in (
        'login',
        'activate_two_factor_authentication_method',
        'deactivate_two_factor_authentication_method'
    ):
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if reason in ('login', ) and flask_login.current_user.is_authenticated != is_for_refresh:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if reason in (
        'activate_two_factor_authentication_method',
        'deactivate_two_factor_authentication_method'
    ) and not flask_login.current_user.is_authenticated:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    all_methods = authentication.get_two_factor_authentication_methods(
        user_id=user.id,
        active=reason in (
            'login',
            'deactivate_two_factor_authentication_method'
        )
    )
    try:
        method_id = int(int(flask.request.args.get('method_id', '')))
    except ValueError:
        method_id = -1
    method = {
        method.id: method
        for method in all_methods
    }.get(method_id)
    if method is None:
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if reason in (
        'deactivate_two_factor_authentication_method',
    ):
        method_to_deactivate_id = flask.session['confirm_data']['method_to_deactivate_id']
        method_to_deactivate = {
            method.id: method
            for method in all_methods
        }.get(method_to_deactivate_id)
        if method_to_deactivate is None:
            flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
            flask.session.pop('confirm_data')
            return flask.redirect(flask.url_for('.index'))
        if not method_to_deactivate.active:
            flask.flash(_('The two-factor authentication method has already been disabled.'), 'success')
            return flask.redirect(flask.url_for('.user_me_preferences'))
    else:
        method_to_deactivate = None

    if method.data.get('type') != 'fido2_passkey':
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    if not method.data.get('credential_data'):
        flask.flash(_('This two-factor authentication attempt has failed. Please try again.'), 'error')
        flask.session.pop('confirm_data')
        return flask.redirect(flask.url_for('.index'))

    confirm_form = WebAuthnLoginForm()
    base64_credential_data = method.data['credential_data']
    credential_data = AttestedCredentialData(base64.b64decode(base64_credential_data.encode('utf-8')))
    all_credentials = [credential_data]
    server = authentication.get_webauthn_server()
    if confirm_form.validate_on_submit():
        credential = server.authenticate_complete(
            flask.session.pop("webauthn_state"),
            credentials=all_credentials,
            response=json.loads(confirm_form.assertion.data),
        )
        if credential.credential_id == credential_data.credential_id:
            flask.session.pop('confirm_data')
            if reason == 'login':
                return complete_sign_in(user, is_for_refresh, remember_me, shared_device, next_url=next_url)
            elif reason == 'activate_two_factor_authentication_method':
                authentication.activate_two_factor_authentication_method(method.id)
                flask.flash(_('The two-factor authentication method has been enabled.'), 'success')
                return flask.redirect(flask.url_for('.user_me_preferences'))
            elif reason == 'deactivate_two_factor_authentication_method':
                assert method_to_deactivate is not None
                authentication.deactivate_two_factor_authentication_method(method_to_deactivate.id)
                flask.flash(_('The two-factor authentication method has been disabled.'), 'success')
                return flask.redirect(flask.url_for('.user_me_preferences'))
            else:
                return flask.abort(500)
        else:
            flask.flash(_('The passkey does not match the one stored for this two-factor authentication method, please try again.'), 'error')

    options, state = server.authenticate_begin(
        credentials=all_credentials
    )
    flask.session["webauthn_state"] = state
    return flask.render_template(
        'two_factor_authentication/confirm_fido2_passkey.html',
        confirm_form=confirm_form,
        options=dict(options),
        method_id=method_id,
        user=user,
        description=method.data.get('description'),
        reason=reason,
        method_to_deactivate=method_to_deactivate,
    )



# coding: utf-8
"""

"""
import copy
import dataclasses
import datetime
import typing
from hashlib import sha256

import flask
import flask_login
from flask_babel import gettext

from .components import get_component, get_components, Component, check_component_exists
from .. import db, logic
from . import errors, settings
from .utils import cache
from .notifications import create_notification_for_being_automatically_linked
from .. models import users, UserType, FederatedIdentity, Authentication, AuthenticationType


@dataclasses.dataclass(frozen=True)
class UserInvitation:
    """
    This class provides an immutable wrapper around models.users.UserInvitation.
    """
    id: int
    inviter_id: int
    utc_datetime: datetime.datetime
    accepted: bool

    @classmethod
    def from_database(cls, user_invitation: users.UserInvitation) -> 'UserInvitation':
        return UserInvitation(
            id=user_invitation.id,
            inviter_id=user_invitation.inviter_id,
            utc_datetime=user_invitation.utc_datetime,
            accepted=user_invitation.accepted
        )

    @property
    def expired(self) -> bool:
        expiration_datetime = self.utc_datetime + datetime.timedelta(seconds=flask.current_app.config['INVITATION_TIME_LIMIT'])
        return bool(datetime.datetime.now(datetime.timezone.utc) >= expiration_datetime)


@dataclasses.dataclass(frozen=True)
class User:
    """
    This class provides an immutable wrapper around models.users.User.
    """

    id: int
    name: typing.Optional[str]
    email: typing.Optional[str]
    type: UserType
    is_admin: bool
    is_readonly: bool
    is_hidden: bool
    is_active: bool
    orcid: typing.Optional[str]
    affiliation: typing.Optional[str]
    role: typing.Optional[str]
    extra_fields: typing.Dict[str, typing.Any] = dataclasses.field(compare=False)
    fed_id: typing.Optional[int]
    component_id: typing.Optional[int]
    last_modified: datetime.datetime
    last_modified_by_id: typing.Optional[int]
    eln_import_id: typing.Optional[int]
    eln_object_id: typing.Optional[str]
    # for use by .languages.get_user_language, no type hint to avoid circular import
    language_cache: typing.List[typing.Optional[typing.Any]] = dataclasses.field(default_factory=lambda: [None], kw_only=True, repr=False, compare=False)
    # for use by the settings property
    _settings_cache: typing.List[typing.Optional[typing.Dict[str, typing.Any]]] = dataclasses.field(default_factory=lambda: [None], kw_only=True, repr=False, compare=False)

    @classmethod
    def from_database(cls, user: users.User) -> 'User':
        return User(
            id=user.id,
            name=user.name,
            email=user.email,
            type=user.type,
            is_admin=user.is_admin,
            is_readonly=user.is_readonly,
            is_hidden=user.is_hidden,
            is_active=user.is_active,
            orcid=user.orcid,
            affiliation=user.affiliation,
            role=user.role,
            extra_fields=copy.deepcopy(user.extra_fields),
            fed_id=user.fed_id,
            component_id=user.component_id,
            last_modified=user.last_modified,
            last_modified_by_id=user.last_modified_by_id,
            eln_import_id=user.eln_import_id,
            eln_object_id=user.eln_object_id,
        )

    @property
    def component(self) -> typing.Optional[Component]:
        if self.component_id is None:
            return None
        return get_component(self.component_id)

    @property
    def eln_import(self) -> typing.Optional['logic.eln_import.ELNImport']:
        if self.eln_import_id is None:
            return None

        return logic.eln_import.get_eln_import(self.eln_import_id)

    def get_id(self) -> int:
        return self.id

    @property
    def is_authenticated(self) -> bool:
        return self.is_active

    @property
    def is_anonymous(self) -> bool:
        return False

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, (flask_login.UserMixin, User)):
            return bool(self.get_id() == other.get_id())
        return NotImplemented

    def __ne__(self, other: typing.Any) -> bool:
        equal = self.__eq__(other)
        if equal is NotImplemented:
            return NotImplemented
        return not equal

    def get_name(self, include_ref: bool = False, include_id: bool = True, use_local_identity: bool = False) -> str:
        if include_ref and self.component_id is not None:
            db_ref = f', #{self.fed_id} @ {typing.cast(Component, self.component).get_name()}'
        elif include_ref and self.eln_import_id is not None and self.eln_object_id is not None:
            db_ref = f', {self.eln_object_id} @ {gettext("ELN Import")} #{self.eln_import_id}'
        else:
            db_ref = ''
        if use_local_identity:
            local_user = get_user_by_federated_user(federated_user_id=self.id)
        else:
            local_user = None
        if not include_id:
            if db_ref:
                db_ref = " (" + db_ref[2:] + ")"
            if local_user is None or local_user.name is None:
                if self.name is None:
                    return gettext('Imported User') + db_ref
                else:
                    return self.name + db_ref
            else:
                return local_user.name + db_ref
        user_id = local_user.id if local_user else self.id
        name = local_user.name if local_user else self.name
        if name is None:
            if self.component_id or self.eln_import_id:
                name = gettext('Imported User')
            else:
                name = gettext('User')
        return f'{name} (#{user_id}{db_ref})'

    @property
    def has_admin_permissions(self) -> bool:
        """
        Return whether the user has administrator permissions.

        This is the case if the user is an admin and has admin permissions enabled in their settings.
        """
        if not self.is_admin:
            return False
        return bool(self.settings['USE_ADMIN_PERMISSIONS'])

    @property
    def timezone(self) -> typing.Optional[str]:
        if flask.current_app.config['TIMEZONE']:
            return typing.cast(str, flask.current_app.config['TIMEZONE'])
        return typing.cast(typing.Optional[str], self.settings['TIMEZONE'])

    @property
    def settings(self) -> typing.Dict[str, typing.Any]:
        if self._settings_cache[0] is not None:
            return self._settings_cache[0]
        user_settings = settings.get_user_settings(self.id)
        self._settings_cache[0] = user_settings
        return user_settings

    def clear_caches(self) -> None:
        self.language_cache[0] = None
        self._settings_cache[0] = None


class AnonymousUser(flask_login.AnonymousUserMixin):
    @property
    def id(self) -> typing.Optional[int]:
        return None

    @property
    def has_admin_permissions(self) -> bool:
        return False

    @property
    def timezone(self) -> typing.Optional[str]:
        if flask.current_app.config['TIMEZONE']:
            return typing.cast(str, flask.current_app.config['TIMEZONE'])
        return None

    @property
    def is_readonly(self) -> bool:
        # anonymous users cannot change anything but are not specifically marked as readonly
        return False

    @property
    def settings(self) -> typing.Dict[str, typing.Any]:
        return copy.deepcopy(settings.DEFAULT_SETTINGS)


@dataclasses.dataclass(frozen=True)
class UserFederationAlias:
    """
    This class provides an immutable wrapper around models.users.UserFederationAlias.
    """
    user_id: int
    component_id: int
    name: typing.Optional[str]
    use_real_name: bool
    email: typing.Optional[str]
    use_real_email: bool
    orcid: typing.Optional[str]
    use_real_orcid: bool
    affiliation: typing.Optional[str]
    use_real_affiliation: bool
    role: typing.Optional[str]
    use_real_role: bool
    extra_fields: typing.Dict[str, typing.Any]
    last_modified: datetime.datetime
    is_default: bool

    @classmethod
    def from_database(cls, alias: users.UserFederationAlias) -> 'UserFederationAlias':
        if any([alias.use_real_name, alias.use_real_email, alias.use_real_orcid, alias.use_real_affiliation, alias.use_real_role]):
            user = get_user(alias.user_id)
            return UserFederationAlias(
                is_default=False,
                user_id=alias.user_id,
                component_id=alias.component_id,
                name=user.name if alias.use_real_name else alias.name,
                use_real_name=alias.use_real_name,
                email=user.email if alias.use_real_email else alias.email,
                use_real_email=alias.use_real_email,
                orcid=user.orcid if alias.use_real_orcid else alias.orcid,
                use_real_orcid=alias.use_real_orcid,
                affiliation=user.affiliation if alias.use_real_affiliation else alias.affiliation,
                use_real_affiliation=alias.use_real_affiliation,
                role=user.role if alias.use_real_role else alias.role,
                use_real_role=alias.use_real_role,
                extra_fields=copy.deepcopy(alias.extra_fields),
                last_modified=alias.last_modified
            )
        else:
            return UserFederationAlias(
                is_default=False,
                user_id=alias.user_id,
                component_id=alias.component_id,
                name=alias.name,
                use_real_name=alias.use_real_name,
                email=alias.email,
                use_real_email=alias.use_real_email,
                orcid=alias.orcid,
                use_real_orcid=alias.use_real_orcid,
                affiliation=alias.affiliation,
                use_real_affiliation=alias.use_real_affiliation,
                role=alias.role,
                use_real_role=alias.use_real_role,
                extra_fields=copy.deepcopy(alias.extra_fields),
                last_modified=alias.last_modified
            )

    @classmethod
    def from_user_profile(cls, user_id: int, component_id: int) -> 'UserFederationAlias':
        user = get_user(user_id)
        return UserFederationAlias(
            is_default=True,
            user_id=user_id,
            component_id=component_id,
            name=user.name,
            use_real_name=True,
            email=user.email,
            use_real_email=True,
            orcid=user.orcid,
            use_real_orcid=True,
            affiliation=user.affiliation,
            use_real_affiliation=True,
            role=user.role,
            use_real_role=True,
            extra_fields=copy.deepcopy(user.extra_fields),
            last_modified=user.last_modified
        )


@cache
def check_user_exists(
        user_id: int
) -> None:
    """
    Check whether a user with the given user ID exists.

    :param user_id: the ID of an existing user
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    if not db.session.query(db.exists().where(users.User.id == user_id)).scalar():
        raise errors.UserDoesNotExistError()


def get_user(user_id: int, component_id: typing.Optional[int] = None) -> User:
    return User.from_database(get_mutable_user(user_id, component_id))


def get_mutable_user(user_id: int, component_id: typing.Optional[int] = None) -> users.User:
    """
    Get the mutable user instance to perform changes in the database on.

    :param user_id: the user ID of an existing user
    :param component_id: the ID of an existing component, or None
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    if component_id is None or component_id == 0:
        user = users.User.query.filter_by(id=user_id).first()
    else:
        user = users.User.query.filter_by(fed_id=user_id, component_id=component_id).first()
    if user is None:
        if component_id is not None:
            check_component_exists(component_id)
        raise errors.UserDoesNotExistError()
    return user


def update_user(
        user_id: int,
        updating_user_id: typing.Optional[int],
        **attributes: typing.Any
) -> None:
    """
    Update one or multiple attributes of an existing user in the database.

    :param user_id: the user ID of an existing user
    :param updating_user_id: the ID of the user causing the update, or None
    :param attributes: a mapping of attribute names and values
    :raise AttributeError: when a non-existing attribute is set
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID or updating user ID exists
    """
    user = get_mutable_user(user_id)
    for name in attributes:
        if name not in {
            'name',
            'email',
            'type',
            'is_admin',
            'is_readonly',
            'is_hidden',
            'is_active',
            'orcid',
            'affiliation',
            'role',
            'extra_fields',
            'fed_id',
            'component_id',
            'eln_import_id',
            'eln_object_id',
        }:
            raise AttributeError(f"'User' object has no attribute '{name}'")

    if updating_user_id is not None:
        # ensure the updating user exists
        check_user_exists(updating_user_id)

        user.last_modified = datetime.datetime.now(datetime.timezone.utc)
        user.last_modified_by_id = updating_user_id

    for name, value in attributes.items():
        setattr(user, name, value)

    db.session.add(user)
    db.session.commit()


def get_users(
        *,
        exclude_hidden: bool = False,
        order_by: typing.Optional[db.Column] = users.User.name,  # type: ignore
        exclude_fed: bool = False,
        exclude_eln_import: bool = False,
) -> typing.List[User]:
    """
    Returns all users.

    :param exclude_hidden: whether to exclude hidden users
    :param order_by: Column to order the users by, or None
    :param exclude_fed: whether to exclude federation users
    :param exclude_eln_import: whether to exclude users from an .eln file
    :return: the list of users
    """
    user_query = users.User.query
    if exclude_hidden:
        user_query = user_query.filter_by(is_hidden=False)
    if order_by is not None:
        user_query = user_query.order_by(order_by)
    if exclude_fed:
        user_query = user_query.filter(users.User.type != UserType.FEDERATION_USER)
    if exclude_eln_import:
        user_query = user_query.filter(users.User.type != UserType.ELN_IMPORT_USER)
    return [
        User.from_database(user)
        for user in user_query.all()
    ]


def get_users_by_email(email: str) -> typing.List[User]:
    return [
        User.from_database(user)
        for user in
        users.User.query.filter_by(email=email).all()
    ]


def get_users_for_component(
        component_id: int,
        exclude_hidden: bool = False,
        order_by: typing.Optional[db.Column] = users.User.name  # type: ignore
) -> typing.List[User]:
    user_query = users.User.query.filter_by(component_id=component_id)
    if exclude_hidden:
        user_query = user_query.filter_by(is_hidden=False)
    if order_by is not None:
        user_query = user_query.order_by(order_by)
    return [
        User.from_database(user)
        for user in user_query.all()
    ]


def get_administrators() -> typing.List[User]:
    """
    Returns all current administrators.

    :return: the list of administrators
    """
    return [
        User.from_database(user)
        for user in users.User.query.filter_by(is_admin=True).all()
    ]


def get_users_by_name(name: str) -> typing.List[User]:
    """
    Return all users with a given name.

    :param name: the user name to search for
    :return: the list of users with this name
    """
    return [
        User.from_database(user)
        for user in users.User.query.filter_by(name=name).all()
    ]


@typing.overload
def create_user(
        name: str,
        email: str,
        type: UserType,
        orcid: typing.Optional[str] = None,
        affiliation: typing.Optional[str] = None,
        role: typing.Optional[str] = None,
        extra_fields: typing.Optional[typing.Dict[str, str]] = None,
        *,
        fed_id: None = None,
        component_id: None = None,
        eln_import_id: None = None,
        eln_object_id: None = None
) -> User:
    ...


@typing.overload
def create_user(
        name: typing.Optional[str],
        email: typing.Optional[str],
        type: UserType,
        orcid: typing.Optional[str] = None,
        affiliation: typing.Optional[str] = None,
        role: typing.Optional[str] = None,
        extra_fields: typing.Optional[typing.Dict[str, str]] = None,
        *,
        fed_id: int,
        component_id: int,
        eln_import_id: None = None,
        eln_object_id: None = None,
) -> User:
    ...


@typing.overload
def create_user(
        name: typing.Optional[str],
        email: typing.Optional[str],
        type: UserType,
        orcid: typing.Optional[str] = None,
        affiliation: typing.Optional[str] = None,
        role: typing.Optional[str] = None,
        extra_fields: typing.Optional[typing.Dict[str, str]] = None,
        *,
        fed_id: None = None,
        component_id: None = None,
        eln_import_id: int,
        eln_object_id: str,
) -> User:
    ...


def create_user(
        name: typing.Optional[str],
        email: typing.Optional[str],
        type: UserType,
        orcid: typing.Optional[str] = None,
        affiliation: typing.Optional[str] = None,
        role: typing.Optional[str] = None,
        extra_fields: typing.Optional[typing.Dict[str, str]] = None,
        *,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None,
        eln_import_id: typing.Optional[int] = None,
        eln_object_id: typing.Optional[str] = None,
) -> User:
    """
    Create a new user.

    This function cannot create a user as an administrator. To set whether a user is
    an administrator, use the set_administrator script or set_user_administrator function.

    :param name: the user's name
    :param email: the user's email address
    :param type: the user's type
    :param orcid: the user's ORCID iD
    :param affiliation: the user's affiliation
    :param role: the user's role
    :param extra_fields: a dict describing the values for defined extra fields
    :param fed_id: the ID of the related user at the exporting component
    :param component_id: the ID of the exporting component
    :param eln_import_id: the ID of the ELN import
    :param eln_object_id: the ID of the user node inside the ELN import
    :return: the newly created user
    """

    assert (component_id is None) == (fed_id is None)
    assert (eln_import_id is None) == (eln_object_id is None)
    assert component_id is not None or eln_import_id is not None or (name is not None and email is not None)

    if component_id is not None:
        check_component_exists(component_id)

    if eln_import_id is not None:
        logic.eln_import.get_eln_import(eln_import_id)

    user = users.User(
        name=name,
        email=email,
        type=type,
        orcid=orcid,
        affiliation=affiliation,
        role=role,
        extra_fields=extra_fields,
        fed_id=fed_id,
        component_id=component_id,
        last_modified=datetime.datetime.now(datetime.timezone.utc),
        eln_import_id=eln_import_id,
        eln_object_id=eln_object_id,
    )
    db.session.add(user)
    db.session.commit()
    return User.from_database(user)


def set_user_readonly(user_id: int, readonly: bool) -> None:
    """
    Set whether a user should be limited to READ permissions.

    :param user_id: the user ID of an existing user
    :param readonly: True, if the user should be read only, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    update_user(
        user_id,
        updating_user_id=None,
        is_readonly=readonly
    )


def set_user_hidden(user_id: int, hidden: bool) -> None:
    """
    Set whether a user should be hidden from user lists.

    :param user_id: the user ID of an existing user
    :param hidden: True, if the user should be hidden, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    update_user(
        user_id,
        updating_user_id=None,
        is_hidden=hidden
    )


def set_user_active(user_id: int, active: bool) -> None:
    """
    Set whether a user should be allowed to sign in.

    :param user_id: the user ID of an existing user
    :param active: True, if the user should be active, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    update_user(
        user_id,
        updating_user_id=None,
        is_active=active
    )


def set_user_administrator(user_id: int, is_admin: bool) -> None:
    """
    Set whether a user is an administrator.

    :param user_id: the user ID of an existing user
    :param is_admin: True, if the user is an administrator, False otherwise
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    update_user(
        user_id,
        updating_user_id=None,
        is_admin=is_admin
    )


def get_user_invitation(invitation_id: int) -> UserInvitation:
    """
    Get an existing invitation.

    :param invitation_id: the ID of an existing invitation
    :return: the invitation
    :raise errors.UserInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = users.UserInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.UserInvitationDoesNotExistError()
    return UserInvitation.from_database(invitation)


def set_user_invitation_accepted(invitation_id: int) -> None:
    """
    Mark an invitation as having been accepted.

    :param invitation_id: the ID of an existing invitation
    :raise errors.UserInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = users.UserInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.UserInvitationDoesNotExistError()
    invitation.accepted = True
    db.session.add(invitation)
    db.session.commit()


def get_user_alias(user_id: int, component_id: int) -> UserFederationAlias:
    """
    Get an existing user alias.

    :param user_id: the ID of an existing user
    :param component_id: the ID of an existing component
    :return: the user alias
    :raise errors.UserDoesNotExistError: when no user with given ID exists
    :raise errors.ComponentDoesNotExistError: when no component with given ID exists
    :raise errors.UserAliasDoesNotExistError: when no alias with given IDs exists
    """
    alias = users.UserFederationAlias.query.filter_by(user_id=user_id, component_id=component_id).first()
    if alias is None:
        user = get_user(user_id)
        check_component_exists(component_id)
        if flask.current_app.config['ENABLE_DEFAULT_USER_ALIASES'] and user.type == UserType.PERSON:
            return UserFederationAlias.from_user_profile(user_id, component_id)
        raise errors.UserAliasDoesNotExistError()
    return UserFederationAlias.from_database(alias)


def get_user_aliases_for_user(user_id: int) -> typing.List[UserFederationAlias]:
    """
    Get all aliases for a user.

    :param user_id: the ID of an existing user
    :return: list of user aliases
    :raise errors.UserDoesNotExistError: when no user with given ID exists
    """
    user = get_user(user_id)
    alias = users.UserFederationAlias.query.filter_by(user_id=user_id).all()
    if flask.current_app.config['ENABLE_DEFAULT_USER_ALIASES'] and user.type == UserType.PERSON:
        return [
            UserFederationAlias.from_database(a) for a in alias
        ] + [
            UserFederationAlias.from_user_profile(user_id, component.id)
            for component in get_components()
            if not any(
                component.id == a.component_id
                for a in alias
            )
        ]
    return [UserFederationAlias.from_database(a) for a in alias]


def get_user_aliases_for_component(component_id: int, modified_since: typing.Optional[datetime.datetime] = None) -> typing.List[UserFederationAlias]:
    """
    Get all aliases for a component.

    :param component_id: the ID of an existing component
    :param modified_since: Only return aliases modified since modified_since. None to query every alias. (default: None)
    :return: list of user aliases
    :raise errors.ComponentDoesNotExistError: when no component with the given ID exists
    """
    check_component_exists(component_id)
    if modified_since is None:
        alias = users.UserFederationAlias.query.filter_by(component_id=component_id).all()
    else:
        alias = users.UserFederationAlias.query.join(
            users.User, users.User.id == users.UserFederationAlias.user_id
        ).filter(
            db.or_(
                db.and_(
                    users.UserFederationAlias.component_id == component_id,
                    users.UserFederationAlias.last_modified >= modified_since
                ),
                db.and_(
                    db.or_(users.UserFederationAlias.use_real_name, users.UserFederationAlias.use_real_email, users.UserFederationAlias.use_real_orcid, users.UserFederationAlias.use_real_affiliation, users.UserFederationAlias.use_real_role),
                    users.User.last_modified >= modified_since
                )
            )
        ).all()
    if flask.current_app.config['ENABLE_DEFAULT_USER_ALIASES']:
        return [
            UserFederationAlias.from_database(a) for a in alias
        ] + [
            UserFederationAlias.from_user_profile(user.id, component_id)
            for user in get_users()
            if not any(
                user.id == a.user_id
                for a in alias
            ) and (modified_since is None or user.last_modified >= modified_since) and user.type == UserType.PERSON
        ]
    return [UserFederationAlias.from_database(a) for a in alias]


def create_user_alias(
    user_id: int,
    component_id: int,
    name: typing.Optional[str] = None,
    use_real_name: bool = False,
    email: typing.Optional[str] = None,
    use_real_email: bool = False,
    orcid: typing.Optional[str] = None,
    use_real_orcid: bool = False,
    affiliation: typing.Optional[str] = None,
    use_real_affiliation: bool = False,
    role: typing.Optional[str] = None,
    use_real_role: bool = False
) -> UserFederationAlias:
    """
    Create a new user alias for a component.

    :param user_id: the ID of an existing user
    :param component_id:  the ID of an existing component
    :param name: the alias name
    :param use_real_name: boolean whether to use the name from the users' profile
    :param email: the alias email
    :param use_real_email: boolean whether to use the email from the users' profile
    :param orcid: the alias orcid
    :param use_real_orcid: boolean whether to use the orcid from the users' profile
    :param affiliation: the alias affiliation
    :param use_real_affiliation: boolean whether to use the affiliation from the users' profile
    :param role: the alias role
    :param use_real_role: boolean whether to use the role from the users' profile
    :return: the new created alias
    :raise errors.UserDoesNotExistError: when no user with given ID exists
    :raise errors.ComponentDoesNotExistError: when no component with given ID exists
    :raise errors.UserAliasAlreadyExistsError: when an alias with given IDs already exists
    """
    alias = users.UserFederationAlias.query.filter_by(user_id=user_id, component_id=component_id).first()
    if alias is not None:
        check_user_exists(user_id)
        check_component_exists(component_id)
        raise errors.UserAliasAlreadyExistsError()
    if use_real_name:
        name = None
    if use_real_email:
        email = None
    if use_real_orcid:
        orcid = None
    if use_real_affiliation:
        affiliation = None
    if use_real_role:
        role = None
    alias = users.UserFederationAlias(user_id, component_id, name, use_real_name, email, use_real_email, orcid, use_real_orcid, affiliation, use_real_affiliation, role, use_real_role, {})
    db.session.add(alias)
    db.session.commit()
    return UserFederationAlias.from_database(alias)


def update_user_alias(
    user_id: int,
    component_id: int,
    name: typing.Optional[str] = None,
    use_real_name: bool = False,
    email: typing.Optional[str] = None,
    use_real_email: bool = False,
    orcid: typing.Optional[str] = None,
    use_real_orcid: bool = False,
    affiliation: typing.Optional[str] = None,
    use_real_affiliation: bool = False,
    role: typing.Optional[str] = None,
    use_real_role: bool = False
) -> None:
    """
    Update a users' alias for a component.

    If a user does not have an alias for this component yet, but default
    aliases are enabled, a new non-default alias will be created with the
    given settings without raising a UserAliasDoesNotExistError exception.

    :param user_id: the ID of an existing user
    :param component_id:  the ID of an existing component
    :param name: the alias name
    :param use_real_name: boolean whether to use the name from the users' profile
    :param email: the alias email
    :param use_real_email: boolean whether to use the email from the users' profile
    :param orcid: the alias orcid
    :param use_real_orcid: boolean whether to use the orcid from the users' profile
    :param affiliation: the alias affiliation
    :param use_real_affiliation: boolean whether to use the affiliation from the users' profile
    :param role: the alias role
    :param use_real_role: boolean whether to use the role from the users' profile
    :raise errors.UserDoesNotExistError: when no user with given ID exists
    :raise errors.ComponentDoesNotExistError: when no component with given ID exists
    :raise errors.UserAliasDoesNotExistError: when no alias with given IDs exists
    """
    alias = users.UserFederationAlias.query.filter_by(user_id=user_id, component_id=component_id).first()
    if alias is None:
        user = get_user(user_id)
        check_component_exists(component_id)
        if flask.current_app.config['ENABLE_DEFAULT_USER_ALIASES'] and user.type == UserType.PERSON:
            create_user_alias(
                user_id=user_id,
                component_id=component_id,
                name=name,
                use_real_name=use_real_name,
                email=email,
                use_real_email=use_real_email,
                orcid=orcid,
                use_real_orcid=use_real_orcid,
                affiliation=affiliation,
                use_real_affiliation=use_real_affiliation,
                role=role,
                use_real_role=use_real_role
            )
            return
        raise errors.UserAliasDoesNotExistError()
    if use_real_name:
        name = None
    if use_real_email:
        email = None
    if use_real_orcid:
        orcid = None
    if use_real_affiliation:
        affiliation = None
    if use_real_role:
        role = None
    alias.name = name
    alias.use_real_name = use_real_name
    alias.email = email
    alias.use_real_email = use_real_email
    alias.orcid = orcid
    alias.use_real_orcid = use_real_orcid
    alias.affiliation = affiliation
    alias.use_real_affiliation = use_real_affiliation
    alias.role = role
    alias.use_real_role = use_real_role
    alias.last_modified = datetime.datetime.now(datetime.timezone.utc)
    db.session.add(alias)
    db.session.commit()


def delete_user_alias(user_id: int, component_id: int) -> None:
    """
    Delete a user alias.

    :param user_id: the ID of an existing user
    :param component_id:  the ID of an existing component
    :raise errors.UserDoesNotExistError: when no user with given ID exists
    :raise errors.ComponentDoesNotExistError: when no component with given ID exists
    :raise errors.UserAliasDoesNotExistError: when no alias with given IDs exists
    """
    alias = users.UserFederationAlias.query.filter_by(user_id=user_id, component_id=component_id).first()
    if alias is None:
        check_user_exists(user_id)
        check_component_exists(component_id)
        raise errors.UserAliasDoesNotExistError()
    db.session.delete(alias)
    db.session.commit()


def create_eln_federated_identity(user_id: int, eln_user_id: int) -> typing.Optional[FederatedIdentity]:
    """
    Create a new federated identity between a user and an eln imported user.

    :param user_id: id of local user
    :param eln_user_id: user id of eln imported user
    :return: not none if federated identity was successfully created
    """
    eln_user = users.User.query.filter_by(id=eln_user_id).first()
    if eln_user is None:
        raise errors.UserDoesNotExistError()
    return _create_federated_identity(
        user_id=user_id,
        eln_or_fed_user=User.from_database(eln_user),
        update_if_inactive=False
    )


def create_sampledb_federated_identity(user_id: int, component: Component, fed_id: int, update_if_inactive: bool = False, use_for_login: bool = False) -> typing.Optional[FederatedIdentity]:
    """
    Create a new federated identity between a user and a federated user.

    :param user_id: id of local user
    :param component: component of the federation
    :param fed_id: federated id of the external user
    :param update_if_inactive: if true, inactive federated identity will be set to active
    :param use_for_login: if true, the federated identity can be used to login via federated login
    :return: not none if federated identity was successfully created
    """
    db_fed_user = users.User.query.filter_by(type=UserType.FEDERATION_USER, component_id=component.id, fed_id=fed_id).first()

    if db_fed_user is None:
        fed_user = create_user(name=None, email=None, type=UserType.FEDERATION_USER, fed_id=fed_id, component_id=component.id)
        set_user_hidden(fed_user.id, hidden=True)
    else:
        fed_user = User.from_database(db_fed_user)

    federated_identity = _create_federated_identity(
        user_id=user_id,
        eln_or_fed_user=fed_user,
        update_if_inactive=update_if_inactive,
        use_for_login=use_for_login
    )

    if federated_identity and use_for_login:
        logic.authentication.add_federated_login_authentication(federated_identity)
    return federated_identity


def _create_federated_identity(user_id: int, eln_or_fed_user: User, update_if_inactive: bool = False, use_for_login: bool = False) -> typing.Optional[FederatedIdentity]:
    """
    Create a new federated user link between a local and federated or eln imported user.

    :param user_id: id of the local user
    :param eln_or_fed_user: external user (federated or eln imported)
    :param update_if_inactive: if true, inactive federated identity will be set to active
    :raises errors.FederatedUserInFederatedIdentity: if the federated user is already in a federated identity
    :raises errors.ELNUserInFederatedIdentity: if the eln user is already in a federated identity
    :return: not none if federated identity was successfully created
    """
    if eln_or_fed_user.type == UserType.FEDERATION_USER:
        identity_query = FederatedIdentity.query.join(users.User, users.User.id == FederatedIdentity.local_fed_id).filter(
            users.User.type == UserType.FEDERATION_USER,
            FederatedIdentity.local_fed_id == eln_or_fed_user.id,
        )

        result = identity_query.filter(FederatedIdentity.active.is_(True)).first()

        if result is not None:
            if result.user_id != user_id:
                raise errors.FederatedUserInFederatedIdentityError()
            else:
                return result
    elif eln_or_fed_user.type == UserType.ELN_IMPORT_USER:
        identity_query = FederatedIdentity.query.join(users.User, users.User.id == FederatedIdentity.local_fed_id).filter(
            users.User.type == UserType.ELN_IMPORT_USER,
            FederatedIdentity.user_id == user_id,
            FederatedIdentity.local_fed_id == eln_or_fed_user.id
        )
        result = identity_query.filter(FederatedIdentity.active.is_(True)).first()

        if result is not None:
            raise errors.ELNUserInFederatedIdentityError()

    inactive_identities = identity_query.filter(FederatedIdentity.active.is_(False)).all()
    if inactive_identities:
        for identity in inactive_identities:
            if identity.user_id == user_id and identity.local_fed_id == eln_or_fed_user.id and not identity.active:
                if not update_if_inactive:
                    return None
                identity.active = True
                db.session.add(identity)
                db.session.commit()
                return identity

    fed_identity = FederatedIdentity(user_id=user_id, local_fed_id=eln_or_fed_user.id, login=use_for_login)
    db.session.add(fed_identity)
    db.session.commit()
    return fed_identity


def get_user_by_federated_user(federated_user_id: int, login: bool = False) -> typing.Optional[User]:
    """
    Get the local user of a federated identity by a federated user if exists.

    :param federated_user_id: local id of the federated user
    :param login: if true, only users will be returned that are enabled for federated login
    :return: local user or none
    """
    query = FederatedIdentity.query.filter_by(local_fed_id=federated_user_id, active=True)
    if login:
        query = query.filter_by(login=True)
    federated_identity = query.first()
    user = None
    if federated_identity:
        try:
            user = get_user(federated_identity.user_id)
        except errors.UserDoesNotExistError:
            pass
    return user


def get_federated_identity_by_eln_user(federated_user_id: int) -> typing.Optional[FederatedIdentity]:
    """
    Get the local user of a federated identity by a federated user if exists.

    :param federated_user_id: local id of the federated user
    :return: local user or none
    """
    return FederatedIdentity.query.join(users.User, users.User.id == FederatedIdentity.local_fed_id).filter(FederatedIdentity.local_fed_id == federated_user_id, users.User.type == UserType.ELN_IMPORT_USER).first()


def get_federated_identities(user_id: typing.Optional[int] = None, component: typing.Optional[Component] = None, eln_import_id: typing.Optional[int] = None, active_status: typing.Optional[bool] = True) -> list[FederatedIdentity]:
    """
    Get the federated identities of a user or all within a component or an eln import as a list of user links.

    :param user_id: id of the user, or none
    :param component: add filter for a component if not none
    :param eln_import_id: add filter for an eln import id if not none
    :param active_status: active state of the federated identity which should be set, or none for both
    :return: list of filtered federated identities of the specified user
    """
    query = FederatedIdentity.query
    if user_id is not None:
        query = query.filter(FederatedIdentity.user_id == user_id)
    if component is not None:
        query = query.join(users.User, users.User.id == FederatedIdentity.local_fed_id).filter(users.User.component_id == component.id)
    elif eln_import_id is not None:
        query = query.join(users.User, users.User.id == FederatedIdentity.local_fed_id).filter(users.User.eln_import_id == eln_import_id)
    if active_status is not None:
        query = query.filter(FederatedIdentity.active.is_(active_status))
    return [row for row in query.all()]


def get_federated_identity(user_id: int, eln_or_fed_user_id: int) -> FederatedIdentity:
    """
    Get the federated identity specified by user_id and the local user id of the federated user.

    :param user_id: id of the user
    :param eln_or_fed_user_id: local id of the federated user
    """
    identity = FederatedIdentity.query.filter_by(user_id=user_id, local_fed_id=eln_or_fed_user_id).first()
    if not identity:
        raise errors.FederatedIdentityNotFoundError()
    return identity


def get_active_federated_identity_by_federated_user(federated_user_id: int) -> typing.Optional[FederatedIdentity]:
    """
    Get the active federated identity of a federated user.

    :param federated_user_id: local id of the federated user
    :return: federated identity of the federated user or none
    """
    return FederatedIdentity.query.filter_by(local_fed_id=federated_user_id, active=True).first()


def get_federated_user_links_by_component_id(component_id: int, only_active: bool = True) -> list[FederatedIdentity]:
    """
    Get all federated user links of a federation component.

    :param component_id: id of the component
    :param only_active: if true, only active federated identities will be returned
    :return: list of federated identities in the component
    """
    query = FederatedIdentity.query.join(users.User, FederatedIdentity.local_fed_id == users.User.id).filter(users.User.component_id == component_id, users.User.type == UserType.FEDERATION_USER)
    if only_active:
        query = query.filter(FederatedIdentity.active.is_(True))
    return [row for row in query.all()]


def get_user_email_hashes(user: User) -> list[str]:
    """
    Get a list of hashed email addresses from the user. Only confirmed email addresses are used.

    :param user: user whose verified email addresses should be used
    :return: all hashed email addresses for the specified user
    """
    confirmed_emails = {user.email}
    authentication_logins = Authentication.query.filter_by(user_id=user.id, type=AuthenticationType.EMAIL, confirmed=True).with_entities(Authentication.login).all()
    confirmed_emails.update({login[0]['login'] for login in authentication_logins})
    return [_hash_credential(str(mail)) for mail in confirmed_emails]


def get_email_hashes_for_federation_candidates(component_id: int) -> dict[str, int]:
    """
    Get a list of candidates within a federation. Each candidate is specified by an email address (hashed) and the
    id of the corresponding user.
    Candidates are users which do not have a federated identity with a user from the specified federation component.

    :param component_id: component in which candidates should be searched
    :return: mapping of hashed email address to corresponding user id
    """
    users = get_users(exclude_fed=True, exclude_eln_import=True)
    federated_user_ids = {federated_user.user_id for federated_user in get_federated_user_links_by_component_id(component_id, only_active=False)}

    result = {_hash_credential(str(user.email)): user.id for user in users if user.id not in federated_user_ids and user.type == UserType.PERSON}
    result.update({
        _hash_credential(auth.login['login']): auth.user_id
        for auth in Authentication.query.filter(Authentication.user_id.not_in(federated_user_ids), Authentication.type == AuthenticationType.EMAIL, Authentication.confirmed.is_(True)).all()
    })

    return result


def link_users_by_email_hashes(component_id: int, user_informations: list[dict[str, typing.Any]]) -> None:
    """
    Compare the hashed email addresses from the `user_informations` with the candidates from `get_email_hashes_for_federation_candidates`.
    If two hashes match, a federated identity will be created.

    :param component_id: id of the component
    :param user_informations: list of dictionaries containing keys: "user_id", "email_hashes", "component_uuid" for each entry
    """
    component = get_component(component_id)

    user_id_by_credential_hashes = get_email_hashes_for_federation_candidates(component_id=component_id)

    component_fed_ids = {user.fed_id for user in users.User.query.join(FederatedIdentity, users.User.id == FederatedIdentity.local_fed_id).filter(users.User.component_id == component_id).all()}

    for user_data in user_informations:
        for email_hash in user_data['email_hashes']:
            user_id = user_id_by_credential_hashes.get(email_hash, None)
            fed_id = user_data['user_id']
            if user_id is not None and fed_id not in component_fed_ids:
                create_sampledb_federated_identity(user_id, component, fed_id, use_for_login=False)
                create_notification_for_being_automatically_linked(
                    user_id=user_id,
                    component_id=component_id,
                    fed_id=fed_id
                )
                break


def delete_user_link(user_id: int, local_fed_user_id: int) -> None:
    """
    Delete the federated identity of a user by the user id and the local id of a federated/imported user.

    :param user_id: id of the local user
    :param local_fed_user_id: local id of the federated user
    """
    identity = FederatedIdentity.query.filter_by(user_id=user_id, local_fed_id=local_fed_user_id).first()
    if identity:
        db.session.delete(identity)
        db.session.commit()


def delete_user_links_by_fed_ids(user_id: int, component: Component, fed_ids: typing.Sequence[int]) -> None:
    """
    Delete the user links of a user within a federation component by the federated ids.

    :param user_id: id of the local user
    :param component: federation component
    :param fed_ids: federation ids of user within the component
    """
    rows = FederatedIdentity.query.join(users.User, FederatedIdentity.local_fed_id == users.User.id).filter(FederatedIdentity.user_id == user_id, users.User.component_id == component.id, users.User.fed_id.in_(fed_ids)).all()
    for row in rows:
        db.session.delete(row)
    db.session.commit()


def revoke_user_link(user_id: int, eln_or_fed_user_id: int) -> None:
    """
    Revoke a federated identity by the user id and the local id of the federated/imported user

    :param user_id: id of the local user
    :param eln_or_fed_user_id: local id of the federated user
    """
    identity = FederatedIdentity.query.filter_by(user_id=user_id, local_fed_id=eln_or_fed_user_id, active=True).first()
    if identity:
        if identity.login:
            auth = Authentication.query.filter(
                Authentication.user_id == identity.user_id,
                Authentication.login['fed_user_id'].astext.cast(db.Integer) == identity.local_fed_user.fed_id,
                Authentication.login['component_id'].astext.cast(db.Integer) == identity.local_fed_user.component_id
            ).first()
            if auth is not None:
                db.session.delete(auth)
        identity.active = False
        db.session.add(identity)
        db.session.commit()


def revoke_user_links_by_fed_ids(user_id: int, component_id: int, fed_ids: typing.Sequence[int]) -> None:
    """
    Revoke the user links of a user within a federation component by the federated ids.

    :user_id: local user which has the federated identity
    :component_id: federation component
    :fed_ids: federated ids of the external users
    """
    base_query = FederatedIdentity.query.join(users.User, FederatedIdentity.local_fed_id == users.User.id).filter(FederatedIdentity.user_id == user_id, users.User.component_id == component_id)
    rows = base_query.filter(users.User.fed_id.in_(fed_ids)).all()

    has_other_auth_methods = identity_not_required_for_auth(component_id)

    remaining_rows_with_login = base_query.filter(users.User.fed_id.not_in(fed_ids), FederatedIdentity.login.is_(True)).all()
    rows_to_delete_with_login = base_query.filter(users.User.fed_id.in_(fed_ids), FederatedIdentity.login.is_(True)).all()

    if not has_other_auth_methods and not remaining_rows_with_login and rows_to_delete_with_login:
        raise errors.NoAuthenticationMethodError()

    for row in rows:
        if row.login:
            auth = Authentication.query.filter(
                Authentication.user_id == row.user_id,
                Authentication.login['fed_user_id'].astext.cast(db.Integer) == row.local_fed_user.fed_id,
                Authentication.login['component_id'].astext.cast(db.Integer) == row.local_fed_user.component_id
            ).first()
            if auth is not None:
                db.session.delete(auth)
        row.active = False
        db.session.add(row)
    db.session.commit()


def enable_user_link(user_id: int, eln_or_fed_user_id: int) -> None:
    """
    Enable a federated identity by the user id and the local id of the federated/imported user

    :param user_id: id of the local user
    :param eln_or_fed_user_id: local id of the federated user
    """
    fed_identity = FederatedIdentity.query.filter_by(user_id=user_id, local_fed_id=eln_or_fed_user_id, active=False).first()
    if fed_identity:
        fed_identity.active = True
        db.session.add(fed_identity)
        db.session.commit()


def enable_federated_identity_for_login(identity: FederatedIdentity) -> None:
    """
    Allow to use the specified federated identity for the federated login

    :identity: federated identity to be enabled for the login
    """
    if not identity.login:
        try:
            logic.authentication.add_federated_login_authentication(federated_identity=identity)
        except errors.AuthenticationMethodAlreadyExists:
            pass
        identity.login = True
        db.session.add(identity)
        db.session.commit()


def identity_not_required_for_auth(component_id: int) -> bool:
    """
    Check if the federated identity of the user with users from the specified federation component is required for authentication.

    :component_id: ID of the federation component
    :return: True if the federated identity is required
    """
    return not flask.current_app.config['ENABLE_FEDERATED_LOGIN'] or (
        Authentication.query.filter(
            db.and_(Authentication.user_id == flask_login.current_user.id, Authentication.confirmed.is_(True), Authentication.type != AuthenticationType.API_TOKEN, Authentication.type != AuthenticationType.API_ACCESS_TOKEN,
                    db.or_(Authentication.type != AuthenticationType.FEDERATED_LOGIN, db.and_(Authentication.type == AuthenticationType.FEDERATED_LOGIN, Authentication.login['component_id'].astext.cast(db.Integer) != component_id)))
        ).count() > 0
    )


def _hash_credential(credential: str) -> str:
    """
    Hash a credential with sha256-algorithm.

    :param credential: credential that should be hashed, e.g. email address
    :return: hash value of the credential
    """
    return sha256(credential.encode()).hexdigest()



# coding: utf-8
"""
Replace description_as_html column with description_is_markdown in actions table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('actions', 'description_is_markdown'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE actions
        ADD description_is_markdown BOOLEAN NOT NULL DEFAULT FALSE
    """))
    db.session.execute(db.text("""
        UPDATE actions
        SET description_is_markdown = TRUE
        WHERE description_as_html IS NOT NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE actions
        DROP description_as_html
    """))
    return True



# coding: utf-8
"""

"""
import json
import typing

import flask
import flask_login
from flask_babel import _

from . import frontend
from .. import logic
from ..models import Permissions, NotificationType
from ..logic.security_tokens import verify_token
from ..logic.languages import get_languages, get_language, get_language_by_lang_code
from ..models.languages import Language
from .projects_forms import CreateProjectForm, EditProjectForm, LeaveProjectForm, InviteUserToProjectForm, InviteGroupToProjectForm, AddSubprojectForm, RemoveSubprojectForm, DeleteProjectForm, RemoveProjectMemberForm, RemoveProjectGroupForm, ObjectLinkForm
from .permission_forms import PermissionsForm
from .utils import check_current_user_is_not_readonly
from ..utils import FlaskResponseT
from ..logic.utils import get_translated_text
from .users_forms import RevokeInvitationForm


@frontend.route('/projects/<int:project_id>', methods=['GET', 'POST'])
@flask_login.login_required
def project(project_id: int) -> FlaskResponseT:
    if 'token' in flask.request.args:
        token = flask.request.args.get('token', '')
        expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
        token_data = verify_token(token, salt='invite_to_project', secret_key=flask.current_app.config['SECRET_KEY'], expiration=expiration_time_limit)
        if token_data is None:
            flask.flash(_('Invalid project group invitation token. Please request a new invitation.'), 'error')
            return flask.abort(403)
        if 'invitation_id' in token_data:
            try:
                project_invitation = logic.projects.get_project_invitation(token_data['invitation_id'])
            except logic.errors.ProjectInvitationDoesNotExistError:
                flask.flash(_('Unknown project group invitation. Please request a new invitation.'), 'error')
                return flask.abort(403)
            if project_invitation.accepted:
                flask.flash(_('This invitation token has already been used. Please request a new invitation.'), 'error')
                return flask.abort(403)
            if project_invitation.revoked:
                flask.flash(_('This invitation has been revoked. Please request a new invitation.'), 'error')
                return flask.abort(403)
        if token_data.get('project_id', None) != project_id:
            return flask.abort(403)
        permissions = Permissions.from_value(token_data.get('permissions', Permissions.READ.value))
        if permissions == Permissions.NONE:
            flask.flash(_('Invalid permissions in project group invitation token. Please request a new invitation.'), 'error')
            return flask.abort(403)
        user_id = token_data.get('user_id', None)
        if user_id != flask_login.current_user.id:
            if user_id is not None:
                try:
                    invited_user = logic.users.get_user(user_id)
                    flask.flash(_('Please sign in as user "%(user_name)s" to accept this invitation.', user_name=invited_user.name), 'error')
                except logic.errors.UserDoesNotExistError:
                    pass
            return flask.abort(403)
        if not flask.current_app.config['DISABLE_SUBPROJECTS']:
            other_project_ids = token_data.get('other_project_ids', [])
            for notification in logic.notifications.get_notifications(user_id, unread_only=True):
                if notification.type == NotificationType.INVITED_TO_PROJECT:
                    if notification.data['project_id'] == project_id:
                        logic.notifications.mark_notification_as_read(notification.id)
        else:
            other_project_ids = []
        try:
            logic.projects.add_user_to_project(project_id, user_id, permissions, other_project_ids=other_project_ids)
        except logic.errors.UserAlreadyMemberOfProjectError:
            flask.flash(_('You are already a member of this project group.'), 'error')
        except logic.errors.ProjectDoesNotExistError:
            pass
    allowed_language_ids = [
        language.id
        for language in get_languages(only_enabled_for_input=True)
    ]
    user_id = flask_login.current_user.id
    try:
        project = logic.projects.get_project(project_id)
    except logic.errors.ProjectDoesNotExistError:
        return flask.abort(404)
    user_permissions = logic.projects.get_user_project_permissions(project_id=project_id, user_id=user_id, include_groups=True)
    if Permissions.READ in user_permissions:
        leave_project_form = LeaveProjectForm()
    else:
        leave_project_form = None
    if flask_login.current_user.has_admin_permissions:
        user_permissions = Permissions.GRANT
    show_objects_link = Permissions.READ in user_permissions

    class ProjectTranslation(typing.TypedDict):
        language_id: typing.Union[int, str]
        lang_name: str
        name: str
        description: str

    translations: typing.List[ProjectTranslation] = []
    name_language_ids = []
    description_language_ids = []
    if Permissions.WRITE in user_permissions:
        edit_project_form = EditProjectForm()
        group_categories = list(logic.group_categories.get_group_categories())
        group_categories.sort(key=lambda category: get_translated_text(category.name).lower())
        edit_project_form.categories.choices = [
            (str(category.id), category)
            for category in group_categories
        ]
        for lang_code, name in project.name.items():
            lang_id = get_language_by_lang_code(lang_code).id
            name_language_ids.append(lang_id)
            translation: ProjectTranslation = {
                'language_id': lang_id,
                'lang_name': get_translated_text(get_language(lang_id).names),
                'name': name,
                'description': ''
            }
            translations.append(translation)

        for lang_code, description in project.description.items():
            if lang_code == '':
                continue
            lang_id = get_language_by_lang_code(lang_code).id
            description_language_ids.append(lang_id)
            for translation in translations:
                if lang_id == translation['language_id']:
                    translation['description'] = description
                    break
            else:
                translation = {
                    'language_id': lang_id,
                    'lang_name': get_translated_text(get_language(lang_id).names),
                    'name': '',
                    'description': description
                }
                translations.append(translation)
    else:
        edit_project_form = None
        group_categories = None
    show_edit_form = False
    english = get_language(Language.ENGLISH)
    if english.id not in description_language_ids:
        description_language_ids.append(english.id)

    project_member_user_ids_and_permissions = logic.projects.get_project_member_user_ids_and_permissions(project_id=project_id, include_groups=False)
    project_member_group_ids_and_permissions = logic.projects.get_project_member_group_ids_and_permissions(project_id=project_id)

    project_member_user_ids = list(project_member_user_ids_and_permissions.keys())
    project_member_user_ids.sort(key=lambda user_id: (logic.users.get_user(user_id).name or '').lower())

    project_member_group_ids = list(project_member_group_ids_and_permissions.keys())
    project_member_group_ids.sort(key=lambda group_id: get_translated_text(logic.groups.get_group(group_id).name).lower())

    if Permissions.GRANT in user_permissions:
        revoke_invitation_form = RevokeInvitationForm()
        if 'revoke_invitation' in flask.request.form and revoke_invitation_form.validate_on_submit():
            invitation_id = revoke_invitation_form.invitation_id.data
            try:
                invitation = logic.projects.get_project_invitation(invitation_id)
            except logic.errors.GroupInvitationDoesNotExistError:
                flask.flash(_('Unknown project group invitation.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            if invitation.accepted:
                flask.flash(_('This project group invitation has already been accepted.'), 'error')
            else:
                logic.projects.revoke_project_invitation(invitation_id)
                flask.flash(_('The invitation has been revoked.'), 'success')
        invitable_user_list = []
        for user in logic.users.get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'], exclude_fed=True, exclude_eln_import=True):
            if user.id not in project_member_user_ids_and_permissions:
                invitable_user_list.append(user)
        parent_projects_with_add_permissions = logic.projects.get_ancestor_project_ids(project_id, only_if_child_can_add_users_to_ancestor=True)
    else:
        revoke_invitation_form = None
        invitable_user_list = []
        parent_projects_with_add_permissions = set()
    if invitable_user_list:
        other_project_ids_data = []
        for parent_project_id in parent_projects_with_add_permissions:
            other_project_ids_data.append({'project_id': parent_project_id})
        invite_user_form = InviteUserToProjectForm(other_project_ids=other_project_ids_data)
    else:
        invite_user_form = None
    if Permissions.GRANT in user_permissions:
        invitable_group_list = []
        for group in logic.groups.get_user_groups(user_id):
            if group.id not in project_member_group_ids_and_permissions:
                invitable_group_list.append(group)
    else:
        invitable_group_list = []
    if invitable_group_list:
        other_project_ids_data = []
        for parent_project_id in parent_projects_with_add_permissions:
            other_project_ids_data.append({'project_id': parent_project_id})
        invite_group_form = InviteGroupToProjectForm(other_project_ids=other_project_ids_data)
    else:
        invite_group_form = None
    child_project_ids = logic.projects.get_child_project_ids(project_id)
    child_project_ids_can_add_to_parent = {child_project_id: logic.projects.can_child_add_users_to_parent_project(parent_project_id=project_id, child_project_id=child_project_id) for child_project_id in child_project_ids}
    parent_project_ids = logic.projects.get_parent_project_ids(project_id)
    add_subproject_form = None
    remove_subproject_form = None
    delete_project_form = None
    remove_project_member_form = None
    remove_project_group_form = None
    addable_projects = []
    addable_project_ids = []
    if Permissions.GRANT in user_permissions:
        for other_project in logic.projects.get_user_projects(flask_login.current_user.id, include_groups=True):
            if other_project.id == project.id:
                continue
            if Permissions.GRANT in logic.projects.get_user_project_permissions(other_project.id, flask_login.current_user.id, include_groups=True):
                addable_projects.append(other_project)
        addable_project_ids = logic.projects.filter_child_project_candidates(project_id, [child_project.id for child_project in addable_projects])
        addable_projects = [logic.projects.get_project(child_project_id) for child_project_id in addable_project_ids]
        if addable_projects:
            add_subproject_form = AddSubprojectForm()
        if child_project_ids:
            remove_subproject_form = RemoveSubprojectForm()
        if flask_login.current_user.is_admin or not flask.current_app.config['ONLY_ADMINS_CAN_DELETE_PROJECTS'] or (
                set(project_member_user_ids) == {flask_login.current_user.id} and
                not project_member_group_ids
        ) or 'remove_all_permissions' in flask.request.form:
            delete_project_form = DeleteProjectForm()
        remove_project_member_form = RemoveProjectMemberForm()
        remove_project_group_form = RemoveProjectGroupForm()

    project_invitations = None
    show_invitation_log = flask_login.current_user.is_admin and logic.settings.get_user_setting(flask_login.current_user.id, 'SHOW_INVITATION_LOG')
    if Permissions.GRANT in user_permissions or flask_login.current_user.is_admin:
        project_invitations = logic.projects.get_project_invitations(
            project_id=project_id,
            include_accepted_invitations=show_invitation_log,
            include_expired_invitations=show_invitation_log
        )

    object = logic.projects.get_object_linked_to_project(project_id)
    if 'leave' in flask.request.form and Permissions.READ in user_permissions and leave_project_form is not None:
        if leave_project_form.validate_on_submit():
            try:
                logic.projects.remove_user_from_project(project_id=project_id, user_id=user_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.UserDoesNotExistError:
                return flask.abort(500)
            except logic.errors.UserNotMemberOfProjectError:
                flask.flash(_('You have already left the project group.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                flask.flash(_('You cannot leave this project group, because your are the only user with GRANT permissions.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                flask.flash(_('You have successfully left the project group.'), 'success')
                # create object log entry if this caused the deletion of a project linked to an object
                try:
                    logic.projects.get_project(project_id)
                except logic.errors.ProjectDoesNotExistError:
                    if object is not None:
                        logic.object_log.unlink_project(
                            flask_login.current_user.id,
                            object.id,
                            project_id,
                            project_deleted=True
                        )
                return flask.redirect(flask.url_for('.projects'))
    if 'delete' in flask.request.form and Permissions.GRANT in user_permissions and delete_project_form is not None:
        if delete_project_form.validate_on_submit():
            check_current_user_is_not_readonly()
            # create object log entry if deleting a project linked to an object
            if object is not None:
                logic.object_log.unlink_project(
                    flask_login.current_user.id,
                    object.id,
                    project_id,
                    project_deleted=True
                )
            try:
                logic.projects.delete_project(project_id=project_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group has already been deleted.'), 'success')
                return flask.redirect(flask.url_for('.projects'))
            else:
                flask.flash(_('You have successfully deleted the project group.'), 'success')
                return flask.redirect(flask.url_for('.projects'))
    if 'remove_member' in flask.request.form and Permissions.GRANT in user_permissions and remove_project_member_form is not None:
        if remove_project_member_form.validate_on_submit():
            check_current_user_is_not_readonly()
            member_id_str = flask.request.form['remove_member']
            try:
                member_id = int(member_id_str)
            except ValueError:
                flask.flash(_('The member ID was invalid. Please contact an administrator.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            if member_id == flask_login.current_user.id:
                flask.flash(_('You cannot remove yourself from a project group. Please press "Leave Project Group" instead.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                logic.projects.remove_user_from_project(project_id=project_id, user_id=member_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.UserDoesNotExistError:
                flask.flash(_('This user does not exist.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.UserNotMemberOfProjectError:
                flask.flash(_('This user is not a member of this project group.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                flask.flash(_('You cannot remove this user from this project group, because they are the only user with GRANT permissions.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                flask.flash(_('You have successfully removed this user from the project group.'), 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'remove_group' in flask.request.form and Permissions.GRANT in user_permissions and remove_project_group_form is not None:
        if remove_project_group_form.validate_on_submit():
            check_current_user_is_not_readonly()
            group_id_str = flask.request.form['remove_group']
            try:
                group_id = int(group_id_str)
            except ValueError:
                flask.flash(_('The basic group ID was invalid. Please contact an administrator.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                logic.projects.remove_group_from_project(project_id=project_id, group_id=group_id)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.GroupDoesNotExistError:
                flask.flash(_('This basic group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.GroupNotMemberOfProjectError:
                flask.flash(_('This basic group is not a member of this project group.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                flask.flash(_('You cannot remove this basic group from this project group, because they are the only basic group with GRANT permissions.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                flask.flash(_('You have successfully removed this basic group from the project group.'), 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'edit' in flask.request.form and Permissions.WRITE in user_permissions and edit_project_form is not None:
        show_edit_form = True
        if edit_project_form.validate_on_submit():
            check_current_user_is_not_readonly()
            try:
                translations = json.loads(edit_project_form.translations.data)
                if not translations:
                    raise ValueError(_('Please enter at least an english name.'))
                names = {}
                descriptions = {}
                for translation in translations:
                    name = translation['name'].strip()
                    description = translation['description'].strip()
                    language_id = int(translation['language_id'])

                    if language_id not in allowed_language_ids:
                        continue

                    if language_id == Language.ENGLISH:
                        if name == '':
                            raise ValueError(_('Please enter at least an english name.'))
                    elif name == '' and description == '':
                        continue

                    lang_code = get_language(language_id).lang_code

                    names[lang_code] = name
                    if description != '':
                        descriptions[lang_code] = description
                    else:
                        descriptions[lang_code] = ''

                logic.projects.update_project(project_id, names, descriptions)
                logic.group_categories.set_project_group_categories(
                    project_group_id=project_id,
                    category_ids=[
                        int(category_id)
                        for category_id in edit_project_form.categories.data
                    ]
                )
            except ValueError as e:
                flask.flash(str(e), 'error')
                edit_project_form.translations.errors.append(str(e))
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.ProjectAlreadyExistsError:
                edit_project_form.translations.errors.append(_('A project group with this name already exists.'))
            except logic.errors.InvalidProjectNameError:
                edit_project_form.translations.errors.append(_('This project group name is invalid.'))
            else:
                flask.flash(_('Project group information updated successfully.'), 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'add_user' in flask.request.form and Permissions.GRANT in user_permissions and invite_user_form is not None:
        if invite_user_form.validate_on_submit():
            check_current_user_is_not_readonly()
            if not any(user.id == invite_user_form.user_id.data for user in invitable_user_list):
                flask.flash(_('You cannot add this user.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            permissions = Permissions.from_value(invite_user_form.permissions.data)
            if Permissions.READ not in permissions:
                flask.flash(_('Please select read permissions (or higher) for the invitation.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            try:
                other_project_ids = []
                for other_project_id_form in invite_user_form.other_project_ids:
                    try:
                        if other_project_id_form.add_user.data:
                            other_project_ids.append(int(other_project_id_form.project_id.data))
                    except (KeyError, ValueError):
                        pass
                if invite_user_form.add_directly.data and flask_login.current_user.is_admin:
                    logic.projects.add_user_to_project(
                        project_id=project_id,
                        user_id=invite_user_form.user_id.data,
                        other_project_ids=other_project_ids,
                        permissions=permissions
                    )
                    flask.flash(_('The user was successfully added to the project group.'), 'success')
                else:
                    logic.projects.invite_user_to_project(
                        project_id=project_id,
                        user_id=invite_user_form.user_id.data,
                        inviter_id=flask_login.current_user.id,
                        add_to_parent_project_ids=other_project_ids,
                        permissions=permissions
                    )
                    flask.flash(_('The user was successfully invited to the project group.'), 'success')
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.UserDoesNotExistError:
                flask.flash(_('This user does not exist.'), 'error')
            except logic.errors.UserAlreadyMemberOfProjectError:
                flask.flash(_('This user is already a member of this project group.'), 'error')
            else:
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if 'add_group' in flask.request.form and Permissions.GRANT in user_permissions and invite_group_form is not None:
        if invite_group_form.validate_on_submit():
            check_current_user_is_not_readonly()
            if not any(group.id == invite_group_form.group_id.data for group in invitable_group_list):
                flask.flash(_('You cannot add this basic group.'), 'error')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
            try:
                logic.projects.add_group_to_project(project_id, invite_group_form.group_id.data, permissions=Permissions.READ)
            except logic.errors.ProjectDoesNotExistError:
                flask.flash(_('This project group does not exist.'), 'error')
                return flask.redirect(flask.url_for('.projects'))
            except logic.errors.GroupDoesNotExistError:
                flask.flash(_('This basic group does not exist.'), 'error')
            except logic.errors.GroupAlreadyMemberOfProjectError:
                flask.flash(_('This basic group is already a member of this project group.'), 'error')
            else:
                flask.flash(_('The basic group was successfully added to the project group.'), 'success')
                return flask.redirect(flask.url_for('.project', project_id=project_id))
    if not flask.current_app.config['DISABLE_SUBPROJECTS']:
        if 'remove_subproject' in flask.request.form and Permissions.GRANT in user_permissions:
            if remove_subproject_form is not None and remove_subproject_form.validate_on_submit():
                check_current_user_is_not_readonly()
                child_project_id = remove_subproject_form.child_project_id.data
                if child_project_id not in child_project_ids:
                    flask.flash(_('Project group #%(child_project_id)s is not a child of this project group.', child_project_id=int(child_project_id)), 'error')
                else:
                    logic.projects.delete_subproject_relationship(project_id, child_project_id)
                    flask.flash(_('The child project group was successfully removed from this project group.'), 'success')
                    return flask.redirect(flask.url_for('.project', project_id=project_id))
        if 'add_subproject' in flask.request.form and Permissions.GRANT in user_permissions:
            if add_subproject_form is not None and add_subproject_form.validate_on_submit():
                check_current_user_is_not_readonly()
                child_project_id = add_subproject_form.child_project_id.data
                if child_project_id not in addable_project_ids:
                    flask.flash(_('Project group #%(child_project_id)s cannot become a child of this project group.', child_project_id=int(child_project_id)), 'error')
                else:
                    logic.projects.create_subproject_relationship(project_id, child_project_id, child_can_add_users_to_parent=add_subproject_form.child_can_add_users_to_parent.data)
                    flask.flash(_('The child project group was successfully added to this project group.'), 'success')
                    return flask.redirect(flask.url_for('.project', project_id=project_id))
    object_id = object.id if object is not None else None
    object_action = None
    object_link_form = None
    linkable_action_ids = []
    already_linked_object_ids = []
    if Permissions.GRANT in user_permissions and not flask_login.current_user.is_readonly:
        object_link_form = ObjectLinkForm()
        if object is None:
            already_linked_object_ids = [link[1] for link in logic.projects.get_project_object_links()]
            for action_type in logic.action_types.get_action_types():
                if action_type.enable_project_link:
                    linkable_action_ids.extend([
                        action.id
                        for action in logic.actions.get_actions(action_type_id=action_type.id)
                    ])
    if object is not None:
        object_permissions = logic.object_permissions.get_user_object_permissions(object.object_id, flask_login.current_user.id)
        if Permissions.READ in object_permissions and object.action_id is not None:
            object_action = logic.actions.get_action(object.action_id)
        else:
            object = None
    project_group_categories = logic.group_categories.get_project_group_categories(project.id)
    group_category_names = logic.group_categories.get_full_group_category_names()
    return flask.render_template(
        'projects/project.html',
        ENGLISH=english,
        translations=translations,
        languages=get_languages(only_enabled_for_input=True),
        name_language_ids=name_language_ids,
        description_language_ids=description_language_ids,
        get_user=logic.users.get_user,
        get_group=logic.groups.get_group,
        get_project=logic.projects.get_project,
        project=project,
        project_member_user_ids=project_member_user_ids,
        project_member_group_ids=project_member_group_ids,
        project_member_user_ids_and_permissions=project_member_user_ids_and_permissions,
        project_member_group_ids_and_permissions=project_member_group_ids_and_permissions,
        project_invitations=project_invitations,
        project_group_categories=project_group_categories,
        group_categories=group_categories,
        group_category_names=group_category_names,
        show_invitation_log=show_invitation_log,
        object=object,
        object_id=object_id,
        object_link_form=object_link_form,
        linkable_action_ids=linkable_action_ids,
        already_linked_object_ids=already_linked_object_ids,
        object_action=object_action,
        leave_project_form=leave_project_form,
        delete_project_form=delete_project_form,
        remove_project_member_form=remove_project_member_form,
        remove_project_group_form=remove_project_group_form,
        edit_project_form=edit_project_form,
        show_edit_form=show_edit_form,
        invite_user_form=invite_user_form,
        invitable_user_list=invitable_user_list,
        invite_group_form=invite_group_form,
        invitable_group_list=invitable_group_list,
        show_objects_link=show_objects_link,
        child_project_ids=child_project_ids,
        child_project_ids_can_add_to_parent=child_project_ids_can_add_to_parent,
        parent_project_ids=parent_project_ids,
        add_subproject_form=add_subproject_form,
        addable_projects=addable_projects,
        remove_subproject_form=remove_subproject_form,
        user_may_edit_permissions=Permissions.GRANT in user_permissions,
        revoke_invitation_form=revoke_invitation_form,
    )


@frontend.route('/projects/', methods=['GET', 'POST'])
@flask_login.login_required
def projects() -> FlaskResponseT:
    user_id = None
    if 'user_id' in flask.request.args:
        try:
            user_id = int(flask.request.args['user_id'])
        except ValueError:
            pass
    if user_id is not None:
        if user_id != flask_login.current_user.id and not flask_login.current_user.is_admin:
            return flask.abort(403)
        projects = logic.projects.get_user_projects(user_id)
    else:
        projects = logic.projects.get_projects()
    project_permissions_by_id = {
        project.id: logic.projects.get_user_project_permissions(project_id=project.id, user_id=flask_login.current_user.id, include_groups=True)
        for project in projects
    }
    create_project_form = CreateProjectForm()
    group_categories = list(logic.group_categories.get_group_categories())
    group_categories.sort(key=lambda category: get_translated_text(category.name).lower())
    create_project_form.categories.choices = [
        (str(category.id), category)
        for category in group_categories
    ]
    show_create_form = False
    if 'create' in flask.request.form:
        allowed_language_ids = [
            language.id
            for language in get_languages(only_enabled_for_input=True)
        ]
        show_create_form = True
        if create_project_form.validate_on_submit():
            check_current_user_is_not_readonly()
            if flask_login.current_user.is_admin or not flask.current_app.config['ONLY_ADMINS_CAN_CREATE_PROJECTS']:
                try:
                    translations = json.loads(create_project_form.translations.data)
                    if not translations:
                        raise ValueError(_('Please enter at least an english name.'))
                    names = {}
                    descriptions = {}
                    for translation in translations:
                        name = translation['name'].strip()
                        description = translation['description'].strip()
                        language_id = int(translation['language_id'])

                        if language_id not in allowed_language_ids:
                            continue

                        if language_id == Language.ENGLISH:
                            if name == '':
                                raise ValueError(_('Please enter at least an english name.'))

                        lang_code = get_language(language_id).lang_code

                        names[lang_code] = name
                        if description != '':
                            descriptions[lang_code] = description
                        else:
                            descriptions[lang_code] = ''

                    project_id = logic.projects.create_project(names, descriptions, flask_login.current_user.id).id
                    logic.group_categories.set_project_group_categories(
                        project_group_id=project_id,
                        category_ids=[
                            int(category_id)
                            for category_id in create_project_form.categories.data
                        ]
                    )
                except ValueError as e:
                    flask.flash(str(e), 'error')
                    create_project_form.translations.errors.append(str(e))
                except logic.errors.ProjectAlreadyExistsError:
                    create_project_form.translations.errors.append(_('A project group with this name already exists.'))
                except logic.errors.InvalidProjectNameError:
                    create_project_form.translations.errors.append(_('This project group name is invalid.'))
                else:
                    flask.flash(_('The project group has been created successfully.'), 'success')
                    return flask.redirect(flask.url_for('.project', project_id=project_id))
            else:
                create_project_form.translations.errors.append(_('Only administrators can create project groups.'))

    projects_by_id = {
        project.id: project
        for project in projects
    }
    english = get_language(Language.ENGLISH)
    group_categories_by_id = {
        category.id: category
        for category in group_categories
    }
    group_category_names = logic.group_categories.get_full_group_category_names()
    group_category_tree = logic.group_categories.get_group_category_tree(
        basic_group_ids=set(),
        project_group_ids={project.id for project in projects}
    )
    parent_project_ids = logic.projects.get_all_parent_project_ids()
    projects.sort(key=lambda project: get_translated_text(project.name).lower())
    return flask.render_template(
        "projects/projects.html",
        create_project_form=create_project_form,
        show_create_form=show_create_form,
        Permissions=Permissions,
        projects_by_id=projects_by_id,
        project_permissions_by_id=project_permissions_by_id,
        languages=get_languages(only_enabled_for_input=True),
        ENGLISH=english,
        projects=projects,
        parent_project_ids=parent_project_ids,
        group_categories=group_categories,
        group_categories_by_id=group_categories_by_id,
        group_category_names=group_category_names,
        group_category_tree=group_category_tree,
    )


@frontend.route('/projects/<int:project_id>/permissions')
@flask_login.login_required
def project_permissions(project_id: int) -> FlaskResponseT:
    try:
        project = logic.projects.get_project(project_id)
    except logic.errors.ProjectDoesNotExistError:
        return flask.abort(404)

    user_permissions = logic.projects.get_project_member_user_ids_and_permissions(project_id, include_groups=False)
    group_permissions = logic.projects.get_project_member_group_ids_and_permissions(project_id)
    if Permissions.GRANT in logic.projects.get_user_project_permissions(project_id=project_id, user_id=flask_login.current_user.id, include_groups=True) or flask_login.current_user.has_admin_permissions:
        delete_project_form = DeleteProjectForm()
        user_permission_form_data = []
        for user_id, permissions in sorted(user_permissions.items()):
            if user_id is None:
                continue
            user_permission_form_data.append({'user_id': user_id, 'permissions': permissions.name.lower()})
        group_permission_form_data = []
        for group_id, permissions in sorted(group_permissions.items()):
            if group_id is None:
                continue
            group_permission_form_data.append({'group_id': group_id, 'permissions': permissions.name.lower()})
        permissions_form = PermissionsForm(user_permissions=user_permission_form_data, group_permissions=group_permission_form_data)
        # disable permissions for all users and other projects
        permissions_form.all_user_permissions.choices = [('none', Permissions.NONE)]
        permissions_form.project_permissions.max_entries = 0
    else:
        delete_project_form = None
        permissions_form = None
    return flask.render_template(
        'projects/project_permissions.html',
        project=project,
        delete_project_form=delete_project_form,
        user_permissions=user_permissions,
        group_permissions=group_permissions,
        get_user=logic.users.get_user,
        get_group=logic.groups.get_group,
        Permissions=Permissions,
        permissions_form=permissions_form
    )


@frontend.route('/projects/<int:project_id>/permissions', methods=['POST'])
@flask_login.login_required
def update_project_permissions(project_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    try:
        if Permissions.GRANT not in logic.projects.get_user_project_permissions(project_id, flask_login.current_user.id, include_groups=True) and not flask_login.current_user.has_admin_permissions:
            return flask.abort(403)
    except logic.errors.ProjectDoesNotExistError:
        return flask.abort(404)

    permissions_form = PermissionsForm()
    # disable permissions for all users and other projects
    permissions_form.all_user_permissions.choices = [('none', Permissions.NONE)]
    permissions_form.project_permissions.max_entries = 0
    if 'edit_permissions' in flask.request.form and permissions_form.validate_on_submit():
        # First handle GRANT updates, then others (to prevent temporarily not having a GRANT user)
        for user_permissions_data in sorted(permissions_form.user_permissions.data, key=lambda upd: typing.cast(bool, upd['permissions'] != 'grant')):
            user_id = user_permissions_data['user_id']
            try:
                logic.users.check_user_exists(user_id)
            except logic.errors.UserDoesNotExistError:
                continue
            permissions = Permissions.from_name(user_permissions_data['permissions'])
            try:
                logic.projects.update_user_project_permissions(project_id=project_id, user_id=user_id, permissions=permissions)
            except logic.errors.NoMemberWithGrantPermissionsForProjectError:
                continue
        for group_permissions_data in permissions_form.group_permissions.data:
            group_id = group_permissions_data['group_id']
            try:
                logic.groups.get_group(group_id)
            except logic.errors.GroupDoesNotExistError:
                continue
            permissions = Permissions.from_name(group_permissions_data['permissions'])
            logic.projects.update_group_project_permissions(project_id=project_id, group_id=group_id, permissions=permissions)
        flask.flash(_("Successfully updated project group permissions."), 'success')
    else:
        flask.flash(_("A problem occurred while changing the project group permissions. Please try again."), 'error')
    try:
        logic.projects.get_project(project_id)
    except logic.errors.ProjectDoesNotExistError:
        return flask.redirect(flask.url_for('.projects'))
    return flask.redirect(flask.url_for('.project_permissions', project_id=project_id))


@frontend.route('/projects/<int:project_id>/object_link', methods=['POST'])
@flask_login.login_required
def link_object(project_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object_link_form = ObjectLinkForm()
    if not object_link_form.validate_on_submit():
        flask.flash(_("Missing or invalid object ID."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    object_id = object_link_form.object_id.data
    try:
        if Permissions.GRANT not in logic.projects.get_user_project_permissions(project_id, flask_login.current_user.id, include_groups=True) and not flask_login.current_user.has_admin_permissions:
            flask.flash(_("You do not have GRANT permissions for this project group."), 'error')
            return flask.redirect(flask.url_for('.project', project_id=project_id))
        if Permissions.GRANT not in logic.object_permissions.get_user_object_permissions(object_id, flask_login.current_user.id) and not flask_login.current_user.has_admin_permissions:
            flask.flash(_("You do not have GRANT permissions for this object."), 'error')
            return flask.redirect(flask.url_for('.project', project_id=project_id))
        object = logic.objects.get_object(object_id)
        object_is_valid = False
        if object.action_id is not None:
            action = logic.actions.get_action(object.action_id)
            if action.type_id is not None:
                action_type = logic.action_types.get_action_type(action.type_id)
                object_is_valid = action_type.enable_project_link
        if not object_is_valid:
            flask.flash(_("This object cannot be linked to a project group."), 'error')
            return flask.redirect(flask.url_for('.project', project_id=project_id))
        logic.projects.link_project_and_object(project_id, object_id, flask_login.current_user.id)
    except logic.errors.ProjectObjectLinkAlreadyExistsError:
        flask.flash(_("Project group is already linked to an object or object is already linked to a project group."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    except logic.errors.ProjectDoesNotExistError:
        flask.flash(_("Project group does not exist."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    except logic.errors.ObjectDoesNotExistError:
        flask.flash(_("Object does not exist."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    flask.flash(_("Successfully linked the object to a project group."), 'success')
    return flask.redirect(flask.url_for('.project', project_id=project_id))


@frontend.route('/projects/<int:project_id>/object_unlink', methods=['POST'])
@flask_login.login_required
def unlink_object(project_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object_link_form = ObjectLinkForm()
    if not object_link_form.validate_on_submit():
        flask.flash(_("Missing or invalid object ID."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    object_id = object_link_form.object_id.data
    try:
        if Permissions.GRANT not in logic.projects.get_user_project_permissions(project_id, flask_login.current_user.id, include_groups=True) and not flask_login.current_user.has_admin_permissions:
            flask.flash(_("You do not have GRANT permissions for this project group."), 'error')
            return flask.redirect(flask.url_for('.project', project_id=project_id))
        logic.projects.unlink_project_and_object(project_id, object_id, flask_login.current_user.id)
    except logic.errors.ProjectObjectLinkDoesNotExistsError:
        flask.flash(_("No link exists between this object and project group."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    except logic.errors.ProjectDoesNotExistError:
        flask.flash(_("Project group does not exist."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    except logic.errors.ObjectDoesNotExistError:
        flask.flash(_("Object does not exist."), 'error')
        return flask.redirect(flask.url_for('.project', project_id=project_id))
    flask.flash(_("Successfully unlinked the object and project group."), 'success')
    return flask.redirect(flask.url_for('.project', project_id=project_id))



# coding: utf-8
"""

"""
import flask
import minisign
from flask import url_for

from ..logic import minisign_keys
from . import frontend
from ..utils import FlaskResponseT


@frontend.route('/.well-known/keys.json')
def key_list_json() -> FlaskResponseT:
    key_pairs = minisign_keys.get_key_pairs()
    res = [
        {
            "@context": "https://schema.org",
            "@type": "MediaObject",
            "contentUrl": url_for("frontend.minisign_pub_key_by_id", keypair_id=kp.id, _external=True),
            "name": f"ed25519_pub_{str(kp.id)}",
            "encodingFormat": "application/x-minisign-key",
            "description": "Signing key for exported archives",
            "dateCreated": kp.utc_datetime_created
        }
        for kp in key_pairs
    ]
    return flask.jsonify(res)


@frontend.route('/.well-known/keys/<int:keypair_id>')
def minisign_pub_key_by_id(keypair_id: int) -> FlaskResponseT:
    kp = minisign_keys.get_key_pair_by_id(keypair_id)
    if kp is None:
        return flask.abort(404)
    public_key = minisign.PublicKey.from_bytes(kp.pk_bytes)
    key_bytes: bytes = public_key.to_base64()
    return key_bytes.decode('utf-8')



# coding: utf-8
"""

"""

import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship, Query, Mapped

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .actions import Action
    from .instruments import Instrument
    from .locations import Location


class Topic(Model):
    __tablename__ = 'topics'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    description: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    show_on_frontpage: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    show_in_navbar: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    order_index: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    actions: Mapped[typing.List['Action']] = relationship('Action', secondary='action_topics', back_populates='topics')
    instruments: Mapped[typing.List['Instrument']] = relationship('Instrument', secondary='instrument_topics', back_populates='topics')
    short_description: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    description_is_markdown: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    short_description_is_markdown: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    locations: Mapped[typing.List['Location']] = relationship('Location', secondary='location_topics', back_populates='topics')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Topic"]]

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id!r})>'



import datetime
import typing

from .. import db
from .. import models
from ..models import LocationLogEntryType
from . import errors, users, locations, object_permissions, objects


def get_log_entries_for_location(
        location_id: int,
        user_id: typing.Optional[int]
) -> typing.List[typing.Dict[str, typing.Any]]:
    """
    Return a list of log entries for a given location.

    Log entries are returned as dicts, with log entries of some types
    augmented with auxiliary information, such as a referenced object location
    assignment or object. The given user ID will be used to ensure that only
    information which the user may access will be returned.

    :param location_id: the ID of an existing location
    :param user_id: the ID of an existing user, or None
    :return: the location log entries
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """

    log_entries = []
    users_by_id: typing.Dict[typing.Optional[int], typing.Optional[users.User]] = {None: None}
    locations_by_id = {}
    objects_by_id = {}
    for log_entry in models.LocationLogEntry.query.filter_by(
        location_id=location_id
    ).order_by(
        db.desc(models.LocationLogEntry.utc_datetime)
    ).all():
        log_entries.append({
            'type': log_entry.type,
            'utc_datetime': log_entry.utc_datetime,
            'user_id': log_entry.user_id,
            'location_id': log_entry.location_id,
            'data': log_entry.data
        })
        if log_entry.user_id not in users_by_id and log_entry.user_id is not None:
            try:
                users_by_id[log_entry.user_id] = users.get_user(log_entry.user_id)
            except errors.UserDoesNotExistError:
                users_by_id[log_entry.user_id] = None
        log_entries[-1]['user'] = users_by_id[log_entry.user_id]
        object_location_assignment_id = log_entry.data.get('object_location_assignment_id')
        if object_location_assignment_id:
            try:
                object_location_assignment = locations.get_object_location_assignment(object_location_assignment_id)
            except errors.ObjectLocationAssignmentDoesNotExistError:
                object_location_assignment = None
            log_entries[-1]['object_location_assignment'] = object_location_assignment
            if object_location_assignment is not None:
                object_id = object_location_assignment.object_id
                log_entries[-1]['object_id'] = object_id
                if object_id not in objects_by_id:
                    try:
                        if models.Permissions.READ in object_permissions.get_user_object_permissions(object_id, user_id):
                            object = objects.get_object(object_id)
                        else:
                            object = None
                    except errors.ObjectDoesNotExistError:
                        object = None
                    objects_by_id[object_id] = object
                log_entries[-1]['object'] = objects_by_id[object_id]
                if object_location_assignment.location_id != location_id:
                    new_location_id = object_location_assignment.location_id
                    log_entries[-1]['new_location_id'] = new_location_id
                    if new_location_id not in locations_by_id:
                        new_location = None
                        if new_location_id is not None:
                            try:
                                new_location = locations.get_location(new_location_id)
                            except errors.LocationDoesNotExistError:
                                pass
                        locations_by_id[new_location_id] = new_location
                    log_entries[-1]['new_location'] = locations_by_id[new_location_id]
    if not log_entries:
        # ensure the location exists
        locations.check_location_exists(location_id)
    return log_entries


def _store_new_log_entry(
        type: LocationLogEntryType,
        location_id: int,
        user_id: typing.Optional[int],
        data: typing.Dict[str, typing.Any]
) -> None:
    user_log_entry = models.LocationLogEntry(
        type=type,
        location_id=location_id,
        user_id=user_id,
        data=data,
        utc_datetime=datetime.datetime.now(datetime.timezone.utc)
    )
    db.session.add(user_log_entry)
    db.session.commit()


def create_location(user_id: typing.Optional[int], location_id: int) -> None:
    _store_new_log_entry(
        type=LocationLogEntryType.CREATE_LOCATION,
        location_id=location_id,
        user_id=user_id,
        data={
        }
    )


def update_location(user_id: typing.Optional[int], location_id: int) -> None:
    _store_new_log_entry(
        type=LocationLogEntryType.UPDATE_LOCATION,
        location_id=location_id,
        user_id=user_id,
        data={
        }
    )


def add_object(user_id: typing.Optional[int], location_id: int, object_location_assignment_id: int) -> None:
    _store_new_log_entry(
        type=LocationLogEntryType.ADD_OBJECT,
        location_id=location_id,
        user_id=user_id,
        data={
            'object_location_assignment_id': object_location_assignment_id
        }
    )


def change_object(user_id: typing.Optional[int], location_id: int, object_location_assignment_id: int) -> None:
    _store_new_log_entry(
        type=LocationLogEntryType.CHANGE_OBJECT,
        location_id=location_id,
        user_id=user_id,
        data={
            'object_location_assignment_id': object_location_assignment_id
        }
    )


def remove_object(user_id: typing.Optional[int], location_id: int, object_location_assignment_id: int) -> None:
    _store_new_log_entry(
        type=LocationLogEntryType.REMOVE_OBJECT,
        location_id=location_id,
        user_id=user_id,
        data={
            'object_location_assignment_id': object_location_assignment_id
        }
    )
